/*
 * $Id: console_select.h 365 2020-09-08 13:34:31Z gingko $
 * Copyright (C) 2011 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Slecteur des modes de diagnostic utilisant la console
 *
 *	Ce fichier dfinit, selon le mode de compilation (normal, \p _DEBUG ou \p _DIAGNOSTIC)
 *	les macros qui permettent d'activer ou pas l'ouverture de la console de diagnostic.
 **/

#pragma once

// ----------------------------------------------------------------------------

#ifndef USE_CONSOLE
	/**
	 * \def USE_CONSOLE
	 * Dfinit si on veut afficher la console de debugging.
	 * Valeurs possibles :
	 *   - USE_CONSOLE = 0	Aucune console
	 *   - USE_CONSOLE = 1	Une fentre est ouverte pour afficher les messages
	 *   - USE_CONSOLE = -1	Les messages sont envoys vers \p OutputDebugString
	 * Avec USE_CONSOLE=-1, les messages d'erreur sont envoys vers la fonction
	 * \p OutputDebugString. On peut ensuite les rcuprer avec un debugger ou
	 * bien un logiciel tel que DebugView :
	 * http://www.microsoft.com/TechNet/Sysinternals/Utilities/DebugView.mspx
	 **/
	#ifdef _DEBUG
		#define USE_CONSOLE -1
	#else
		#define USE_CONSOLE 0
	#endif
#else
	#if USE_CONSOLE == 1
		// Pour la documentation de la variable avec Doxygen
		#undef USE_CONSOLE
		#define USE_CONSOLE 1
	#endif
#endif

// ----------------------------------------------------------------------------

#if USE_CONSOLE

	#include "console.h"

	#define myprint		OutputMsg
	#define myvprintf	OutputVF
	#define myprintf	OutputF

	#define FMT_ERRX(msg) ERRX(TEXT("chec ") TEXT(msg) TEXT(", code 0x%08x"))

	/// Macro pour test d'erreur simple portant sur un code retour \p DirectShow
	/// et affichant un message de debugging en cas d'erreur
	#define check_hr(hr, msg) myprintf(CMyP FMT_ERRX(msg), FAILED(hr), hr)

#else // #if USE_CONSOLE

	#define myprint(msg)
	#define myprintf(fmt, ...) false
	#define myvprintf(fmt, args) false

	#define FMT_ERRX(msg)
	#define check_hr(hr, msg) FAILED(hr)

	// Macros parfois utilises hors du contexte de la console :
	#define EOL	TEXT("\n")	/// Caractre "fin de ligne"
	#define TAB	TEXT("\t")	/// Caractre "tabulation"

#endif // #if USE_CONSOLE

#if USE_CONSOLE != 1 && !defined(CONSOLE_WITHOUT_COLORS)
	#define CONSOLE_WITHOUT_COLORS 1
#endif

// ----------------------------------------------------------------------------

// Macros d'assistance  l'insertion conditionnelle de courtes squences de code en mode debugging :
// (utile en particulier s'il s'agit de paramtres supplmentaires de fonctions)
#if USE_CONSOLE
	#define IF_CONSOLE(code) code		///< Code conditionnel excut seulement en compilation avec la console
	#define IF_CONSOLE_CA(code) code ,	///< Code conditionnel excut seulement en compilation avec la console, avec une virgule aprs ("comma after")
	#define IF_CONSOLE_CB(code) , code	///< Code conditionnel excut seulement en compilation avec la console, avec une virgule avant ("comma before")
	#define IF_CONSOLE_WE(c1, c2) c1	///< Code conditionnel excut selon compilation avec la console ou pas ("with else")
#else
	#define IF_CONSOLE(code)			///< Dsactive le code avec la console
	#define IF_CONSOLE_CA(arg)			///< Dsactive le code avec la console
	#define IF_CONSOLE_CB(arg)			///< Dsactive le code avec la console
	#define IF_CONSOLE_WE(c1, c2) c2	///< Code conditionnel excut selon compilation avec la console ou pas ("with else")
#endif

#pragma warning(push)
#pragma warning(disable: 4100)	// C4100 : paramtre formel non rfrenc

// ----------------------------------------------------------------------------

/**
 * Initialisation de la console de debugging.
 * \param[in]	pszTitle	Titre de la console
 * \param[in]	pszLogFile	Nom de fichier journal optionnel (mettre NULL si non dsir)
 **/
inline void startConsole(LPCTSTR pszTitle, LPCTSTR pszLogFile=NULL)
{
 #if USE_CONSOLE == 1
  #ifdef _UNICODE
	startConsole(pszTitle, sm_Console, true, pszLogFile);
  #else
	COPY_STR_ON_STACK(CHAR, pszTmp, pszTitle, nLen);
	CharToOem(pszTmp, pszTmp);

	startConsole(pszTmp, sm_Console, true, pszLogFile);
  #endif
 #elif USE_CONSOLE == -1
	startConsole(pszTitle, sm_Debug);
 #endif // #if USE_CONSOLE==1
}

#pragma warning(pop)

// ----------------------------------------------------------------------------

/**
 * Excution d'une squence de code une fois toutes les \p N_VALUE occurrences.
 *
 * Cette macro permet de limiter le nombre d'excutions d'une macro \p myprintf (ou de n'importe quelle autre
 * instruction) de manire , typiquement, limiter un flot de texte de debugging trop important tout en
 * conservant tout de mme quelques informations sur son excution effective.
 * \param[in]	N_VALUE		Nombre d'occurrences entre les excutions
 * \param[in]	...			Code  excuter
 **/
#define ONE_OUT_OF_N(N_VALUE, ...) \
	{static UINT nCount=0; \
	if ((nCount++ % N_VALUE) == 0) { \
		__VA_ARGS__;} }

// ----------------------------------------------------------------------------
