/*
 * $Id: internet.cpp 396 2020-10-15 12:06:34Z gingko $
 * Copyright (C) 2007 lolo_32 <pouchintv-dev@baysse.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation des fonctionnalits de connexion vers Internet
 *
 *	Ce fichier implmente les fonctionnalits ncessaires  la communication
 *	via Internet.
 *
 *	Il est utilis en particulier pour la rcupration des mises  jour de
 *	l'application par Internet.
 **/

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include "stdafx.h"
#include "internet.h"

/**
 * Initialise les valeurs par dfaut, l'User-Agent, la mthode  GET et le type de fichiers accept
 *
 * \param[in]	pszUagt	"user agent"  indiquer lors de la connexion
 * \param[in]	pszUrl	url  interroger
**/
ConnexionInternet::ConnexionInternet(LPCTSTR pszUagt, LPCTSTR pszUrl) :
	m_bChiffre(false),
	m_methode(HTTP_METHODE_GET),
	m_pszUrl(0),
	m_pszServeur(0),
	m_pszUserAgent(0),
	m_hSession(0),
	m_hConnexion(0),
	m_hRequete(0),
	m_dwStatus(0),
	m_dwTaille(0),
	m_dwRecupere(0),
	m_pszContentType(0)
{
	// Dfinition du "User Agent" (nom de l'application avec version dtaille)
	SetUserAgent(pszUagt);

	if (pszUrl && pszUrl[0])
		SetUrl(pszUrl);
}

/**
 * Destructeur de la classe. Elle fait le mnage:
 * # Ferme tous les handles laisss ou non ouverts pour Internet
 * # Libre les chanes de caractres de la classe
 **/
ConnexionInternet::~ConnexionInternet() {
	// Libre le Handle de la requte
	if (this->m_hRequete)
		InternetCloseHandle(this->m_hRequete);

	// Libre le Handle de connexion
	if (this->m_hConnexion)
		InternetCloseHandle(this->m_hConnexion);

	// Libre le Handle de session
	if (this->m_hSession)
		InternetCloseHandle(this->m_hSession);

	// Libre l'url si allou
	if (this->m_pszUrl)
		free(this->m_pszUrl);

	// Libre le nom du serveur si allou
	if (this->m_pszServeur)
		free(this->m_pszServeur);

	// Libre l'User-Agent
	if (this->m_pszUserAgent)
		free(this->m_pszUserAgent);

	// Libre le Content-type (type de fichier)
	if (this->m_pszContentType)
		free(this->m_pszContentType);
}

/**
 * Demande  tablir une connexion scurise (en HTTPS)
 * \param[in]  chiffre  Si true, la connexion sera tablie en HTTPS
 **/
void ConnexionInternet::SetChiffree(bool chiffre) {
	if (chiffre != this->m_bChiffre) {
		// Avant de mmoriser l'option, on regarde s'il n'y a pas de connexions
		// existantes avec l'ancien tat

		if (this->m_hRequete)
			InternetCloseHandle(this->m_hRequete), this->m_hRequete = 0;

		if (this->m_hConnexion)
			InternetCloseHandle(this->m_hConnexion), this->m_hConnexion = 0;

		this->m_bChiffre = chiffre;
	}
}

/**
 * Copie les informations internes, en librant si besoin est le contenu de destination
 *
 * \param[in]  destination  variable destinataire. Si celle-ci est diffrente de NULL, elle sera efface
 * \param[in]  source       chane de caractres  mmoriser
 **/
void ConnexionInternet::SetInfo(LPTSTR *destination, LPCTSTR source) {
	UINT longueur = strlen_T(source);

	// Efface l'ancien paramtre (si il existe)
	if (*destination)
		free(*destination);

	// Remplace par le nouveau
	(*destination) = (LPTSTR)malloc(sizeof(TCHAR)*(longueur+1));
	strcpy_T((*destination), longueur+1, source);
}

/**
 * Remplace l'User-Agent utilis pour la requte
 *
 * \param[in]  useragent  User-Agent devant tre utilis
 **/
void ConnexionInternet::SetUserAgent(LPCTSTR useragent) {
	SetInfo(&(this->m_pszUserAgent), useragent);
}

/**
 * Indique l'url  utiliser pour passer la requte
 *
 * \param[in]  serveur  Serveur revevant la connexion (ex. www.pouchintv.fr)
 * \param[in]  url      Url demande (ex. /forum/)
 * \retval     0        en cas d'erreur
 **/
int ConnexionInternet::SetUrl(LPCTSTR serveur, LPCTSTR url) {
	SetInfo(&(this->m_pszServeur), serveur);
	SetInfo(&(this->m_pszUrl), url);
	return 1;
}

/**
 * Indique l'url  utiliser pour passer la requte
 *
 * \param[in]  url      Url complte demande (ex. http://www.pouchintv.fr/forum/)
 * \retval     0        en cas d'erreur
 **/
int ConnexionInternet::SetUrl(LPCTSTR url) {

	// on ne s'interesse qu'au hostname et  l'URL
	URL_COMPONENTS urlc_explose = {
		sizeof(urlc_explose)	// dwStructSize
		// Le reste  0 (implicite)
	};
	urlc_explose.dwHostNameLength = (DWORD)-1;
	urlc_explose.dwUrlPathLength = (DWORD)-1;

	// Explose l'URL passe
	BOOL result = InternetCrackUrl(
		url,				// Url  exploser
		0,					// taille de url
		0,					// les flags ne sont pas utiliss
		&urlc_explose);		// structure recevant l'Url explose

	if (result == TRUE) {
		// Dfinit l'URL
		LPTSTR hostname = (LPTSTR)malloc(sizeof(TCHAR)*(urlc_explose.dwHostNameLength+1));
		_tcsncpy_s(hostname, urlc_explose.dwHostNameLength+1, urlc_explose.lpszHostName, urlc_explose.dwHostNameLength);
		this->SetUrl(hostname, urlc_explose.lpszUrlPath);
		free(hostname);

		// Dfinit le chiffrage (le cas chant)
		SetChiffree( (urlc_explose.nScheme == INTERNET_SCHEME_HTTPS ? true : false) );
	}
	return result;
}

/** Envoie la connexion
 * \param[in]  donnees      chaine de caractres  envoyer avec la requte
 * \warning                 il faut obligatoirement passer la chaine en CHAR, 
 *                          car si passe en WCHAR, les 0 sont aussi transmis
 * \param[in]  donnees_len  longueur de donnees
 * \param[in]  debut        indique la position de lecture facultative (position pour la rcupration d'un morceau)
 * \retval     0            en cas d'erreur
 * \return                  une valeur de HTTP_ERR indiquant le status de la requte
 * \note                    si donnees != NULL, la requte sera envoy en POST
 **/
int ConnexionInternet::EnvoieRequete(LPCSTR donnees, DWORD donnees_len, DWORD debut) {
	LPCTSTR	pszMethode = 0;

	if (! this->m_hConnexion)
		// Pas de connexion, on l'tablit donc
		this->InitConnexion();

	if (!this->m_hConnexion)
		return 0;

	//
	// On efface les donnes de la prcdente connexion (si elle existent)

	// Rinitialise le status
	this->m_dwStatus = 0;
	// Rinitialise la taille du fichier rponse
	this->m_dwTaille = 0;
	// Rinitialise la date
	ZeroMemory(&this->m_date, sizeof(this->m_date));
	// Rinitialise le nombre d'octets rcuprs
	this->m_dwRecupere = 0;
	// Rinitialise le type de fichier
	if (this->m_pszContentType != 0)
		free(this->m_pszContentType);
	this->m_pszContentType = 0;

	// Si une requte existe, la fermer
	if (this->m_hRequete)
		InternetCloseHandle(this->m_hRequete);

	// Teste la mthode utilise pour envoyer la requte.
	switch (this->m_methode) {
		case HTTP_METHODE_GET:
			pszMethode = TEXT("GET");
			break;

		case HTTP_METHODE_POST:
			pszMethode = TEXT("POST");
			break;
	}

	if (donnees)  // n'a de sens que pour POST ou PUT
		pszMethode = TEXT("POST"), this->m_methode = HTTP_METHODE_POST;

	//
	// Finit de prparer la requte
	LPCTSTR accept[2]	= {TEXT("*/*"), NULL};
	DWORD	dwFlags		= INTERNET_FLAG_NO_CACHE_WRITE |	// Ne pas charger ni utiliser le cache
							INTERNET_FLAG_RESYNCHRONIZE | INTERNET_FLAG_RELOAD;

	if (this->m_bChiffre)
		dwFlags |= INTERNET_FLAG_SECURE;	// Drapeaux diffrents en chiffr ou non

	this->m_hRequete = HttpOpenRequest(
		this->m_hConnexion,			// handle de la connexion prcdement tablie
		pszMethode,					// mthode  utiliser pour la requte
		this->m_pszUrl,				// adresse de la page devant traiter la requte
		NULL,						// version du protocole  utiliser (version par dfaut ici)
		NULL,						// valeur du referer  utiliser (aucun ici)
		accept,						// fichiers accepts
		dwFlags,					// drapeaux
		0);							// contexte d'excution (non utilis ici)

	if (!this->m_hRequete)
		return 0;

	if (this->m_bChiffre) {
		dwFlags = 
			// On ignore le CA racine (au cas o il ne serait pas connu du poste)
			// Il y a tout de mme test du nom d'hote et correspondance dans le certificat
			SECURITY_FLAG_IGNORE_UNKNOWN_CA |
			// Ignore si le certificat est prim
			SECURITY_FLAG_IGNORE_CERT_DATE_INVALID;

		// Sauve les nouvelles options
		InternetSetOption (this->m_hRequete, INTERNET_OPTION_SECURITY_FLAGS, &dwFlags, sizeof (dwFlags) );
	}
	// Ajoute un entte si des donnes sont spcifies en plus
	// c'est obligatoire pour indiquer au serveur qu'il doit attendre de recevoir les donnes 
	// avant de produire la rponse
	if (donnees) {
		LPCTSTR entete = TEXT("Content-Type: application/x-www-form-urlencoded");
		HttpAddRequestHeaders(
			this->m_hRequete,		// handle de la requete auquel ajouter l'entte Range
			entete,					// Entte content-type  rajouter
			47,						// Longueur de l'entte  ajouter
			HTTP_ADDREQ_FLAG_ADD |		// Ajoute l'entte s'il n'existe pas
			HTTP_ADDREQ_FLAG_REPLACE);	// Remplace s'il existe
	}

	// Dtermine si on fait une rcupration partielle
	// Ajoute un en-tte de la forme "Range: bytes=1254725683-"
	if (debut != 0) {
		LPTSTR	entete = 0;
		TCHAR	taille[15];
		_itot_s(debut, taille, _countof(taille), 10);
		// Rcupration partielle
		UINT longueur = 15 /* Zone fixe de l'en tte*/ + strlen_T(taille);
		entete = (LPTSTR)malloc(sizeof(TCHAR)*longueur);

		strcpy_T(entete, longueur, TEXT("Range: bytes="));
		strcat_T(entete, longueur, taille);
		strcat_T(entete, longueur, TEXT("-"));
		HttpAddRequestHeaders(
			this->m_hRequete,		// handle de la requete auquel ajouter l'entte Range
			entete,					// Entte range  rajouter
			longueur,				// longueur de l'entte  ajouter
			HTTP_ADDREQ_FLAG_ADD |		// Ajoute l'entte s'il n'existe pas
			HTTP_ADDREQ_FLAG_REPLACE);	// Remplace s'il existe
		free(entete);
	}

	//
	// Envoie la requte
	BOOL ret = HttpSendRequest(
		this->m_hRequete,			// handle de la requete prpare pour envoie
		0,							// En-tte ajouts  la requte
		0,							// Taille de l'en-tte
		(LPVOID)donnees,			// Donnes  envoyer avec la requte (utilis avec POST, PUT, ...). Si 0, ignor
		donnees_len);				// Taille de donnes

	if (ret != TRUE) {
		DWORD dwErr = GetLastError();

		switch (dwErr) {
			case ERROR_INTERNET_TIMEOUT:
				return HTTP_ERR_TIMEOUT;
			case  ERROR_INTERNET_SEC_CERT_CN_INVALID:
				return HTTP_ERR_CNINVALID;
			case ERROR_INTERNET_SECURITY_CHANNEL_ERROR:
				return HTTP_ERR_SSL;
			default:
				return 0;
		}
	}

	DWORD dwStructsLength = sizeof(DWORD);

	//
	// Rcupre le status de la rponse
	HttpQueryInfo(this->m_hRequete,		// Handle de la requte a vrifier
				HTTP_QUERY_STATUS_CODE |	// Demande  rcuprer le code su status (200 => OK)
				HTTP_QUERY_FLAG_NUMBER,		// sous la forme de DWORD
				&this->m_dwStatus,		// Valeur du status
				&dwStructsLength,		// Taille du tampon
				NULL);					// Inutilis
	//
	// Rcupre la taille du fichier
	HttpQueryInfo(this->m_hRequete,		// Handle de la requte a vrifier
				HTTP_QUERY_CONTENT_LENGTH |	// Demande  rcuprer la taille du ficher
				HTTP_QUERY_FLAG_NUMBER,		// sous la forme de DWORD
				&this->m_dwTaille,		// Taille du fichier
				&dwStructsLength,		// Taille du tampon
				NULL);					// Inutilis
	//
	// Rcupre la date
	dwStructsLength = sizeof(SYSTEMTIME);
	HttpQueryInfo(this->m_hRequete,		// Handle de la requte a vrifier
				HTTP_QUERY_LAST_MODIFIED |	// Date et heure de modification du fichier
				HTTP_QUERY_FLAG_SYSTEMTIME,	// sous la forme de SYSTEMTIME
				&this->m_date,			// Taille du fichier
				&dwStructsLength,		// Taille du tampon
				NULL);					// Inutilis

	return this->m_dwStatus;
}

/**
 * tablie la connection
 *
 * \retval  0  une erreur est survenue
 * \retval  1  tout s'est bien pass
 **/
int ConnexionInternet::InitConnexion() {

	if (! this->m_hSession) {
		// Initialise la session
		this->m_hSession = InternetOpen(
			this->m_pszUserAgent,		// User-agent
			INTERNET_OPEN_TYPE_PRECONFIG, // Utilise le proxy de IE
			NULL,						// Inutile (voir ci-dessus)
			NULL,						// Inutile (voir ci-dessus)
			0);							// Inutilis
	}

	// Initialise la connexion
	this->m_hConnexion = InternetConnect(
		this->m_hSession,				// Handle de session associ  la connexion
		this->m_pszServeur,				// Serveur auquel se connecter
		(this->m_bChiffre ?				// port  utiliser pour se connecter
				INTERNET_DEFAULT_HTTPS_PORT : 
				INTERNET_DEFAULT_HTTP_PORT),		
		NULL,							// nom d'utilisateur de connexion
		NULL,							// mot de passe de connexion
		INTERNET_SERVICE_HTTP,			// service  utiliser (HTTP, FTP ou gopher)

		INTERNET_FLAG_KEEP_CONNECTION |		// Maintenir la connexion ouverte entre 2 requtes
											// TODO: est-ce utile ??????

		INTERNET_FLAG_NO_AUTH |				// Pas d'authentification
		INTERNET_FLAG_NO_UI |				// Pas de fentre pour les cookies
		INTERNET_FLAG_PRAGMA_NOCACHE |		// Demande  ne pas utiliser le serveur proxy
		INTERNET_FLAG_RELOAD,				// Ne prends pas dans le cache local
		0);								// contexte (non utilis)

	if (this->m_hConnexion)
		return 1;
	else
		return 0;
}

/**
 * Lit le contenu de la rponse de la requte prcdente
 *
 * \param[out]    contenu   tableau d'octets du fichiers lu
 * \param[in,out] longueur  longueur de la variable contenue
 * \return                  le nombre d'octets lus
 * \warning                 aucun caractre null ne sera ajout  la fin de contenu
 **/
DWORD ConnexionInternet::RecupereFichier(LPBYTE contenu, DWORD *longueur) {
	if (this->m_dwStatus != HTTP_ERR_OK && this->m_dwStatus != HTTP_ERR_OK_PARTIAL) {
		// Il y a eu une erreur lors de la connexion
		// On arrte le parcour, et indique 0 octets lus
		*longueur = 0;
		return 0;
	}

	DWORD lu = 0;									// Nombre d'octets lus
	DWORD dwOctetsRestant;

	while (InternetQueryDataAvailable(this->m_hRequete, &dwOctetsRestant, 0, 0)) {

		if (lu+dwOctetsRestant > *longueur)
			// Indique le nombre max d'octets pouvant tre lu pour remplir le buffer spcifi
			dwOctetsRestant = *longueur-lu;

		if (dwOctetsRestant == 0)
			// Rien de plus ne peut tre lu
			break;

		DWORD dwOctetsLus;
		BOOL bResult = InternetReadFile(this->m_hRequete, contenu+lu, dwOctetsRestant, &dwOctetsLus);
		this->m_dwRecupere += dwOctetsLus;
		if (!bResult) {
			// Erreur lors de la rcup. Connection interrompue ???
			lu = 0;
			break;
		}

		if (dwOctetsLus == 0)
			break;	// Fin du fichier

		lu += dwOctetsLus;
	}

	// Mmorise le nombre l'octets lue
	*longueur = lu;

	return lu;
}

/** Retourne le type de contenue
 * \param[out]    content  rcupre le type de fichiers (ex. "text/html", "application/x-msdos-program")
 * \param[in,out] longueur la taille de content maximale, et retourne le nombre d'octets copis/necessaires
 **/
void ConnexionInternet::GetContentType(LPTSTR content, UINT *longueur) {
	if (!this->m_pszContentType) {
		// On rcupre le type de fichier

		this->m_pszContentType = (LPTSTR)malloc(sizeof(TCHAR));
		DWORD dwLength = sizeof(TCHAR);

		// Rcupre la taille de l'en-tte
		HttpQueryInfo(this->m_hRequete,		// Handle de la requte a vrifier
					HTTP_QUERY_CONTENT_TYPE,// Type  rcuprer
					this->m_pszContentType,	// Content-type
					&dwLength,				// Taille du tampon (en octet)
					NULL);					// Inutilis

		// Alloue la bonne taille
		free(this->m_pszContentType);
		this->m_pszContentType = (LPTSTR)malloc(dwLength + sizeof(TCHAR));
		dwLength += sizeof(TCHAR);

		// Rcupre les informations, et les stocke
		HttpQueryInfo(this->m_hRequete,		// Handle de la requte a vrifier
					HTTP_QUERY_CONTENT_TYPE,// Type  rcuprer
					this->m_pszContentType,	// Content-type
					&dwLength,				// Taille du tampon
					NULL);					// Inutilis
	} // Fin de la rcupration de l'en-tte

	UINT size = strlen_T(this->m_pszContentType)+1;

	if (*longueur < size) {
		*longueur = size;
		return;
	}
	strcpy_T(content, *longueur, this->m_pszContentType);
	*longueur = size;
}