/*
 * $Id: shared_ptr_ptvm.h 365 2020-09-08 13:34:31Z gingko $
 * Copyright (C) 2010 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation ou liaison vers le template \p shared_ptr.
 *
 *	Ce fichier slectionne les diffrentes sources possibles pour le modle de classe (template)
 *  "shared_ptr", et si aucune n'est disponible, en fournit une implmentation limite.
 *
 *	Le modle "shared_ptr" est normalement inclus dans Visual Studio 2008 dition professionnelle,
 *  mais PAS dans Visual Studio 2005, et pas non plus dans Visual C++ 2008 Express.
 *
 *	L'implmentation limite fournie contient les fonctionnalits minimales ncessaires pour que
 *  le projet PouchinTV Mod puisse tre compil avec Visual Studio 2005, VC++ Express 2005
 *  ou VC++ Express 2008.
 *
 *	Il est galement possible d'utiliser la version qui est incluse dans la bibliothque "Boost",
 *  mais ceci n'a pas t rendu obligatoire afin d'viter d'alourdir les prrequis ncessaires 
 *  la compilation.
 **/

#pragma once

// Si aucun en-tte inclus n'a dfini une quelconque macro ASSERT, en rcuprer une depuis
// la bibliothque CRT
#ifndef ASSERT
	#include <crtdbg.h>
	#define ASSERT _ASSERTE
#endif

#include <memory>
#include "include_config.h"

#if _HAS_TR1 || _MSC_VER >= 1500

	// Si on a une dition de Visual Studio qui inclut l'extension TR1,
	// on utilise la version de "shared_ptr" qui est fournie avec.

	using namespace std;
  #if _MSC_VER < 1900
	using namespace tr1;	// Pour "shared_ptr"
  #endif

#elif EXIST_BOOST

	// Sinon, il est aussi possible d'utiliser la version qui est fournie dans la
	// bibliothque "Boost", si cette bibliothque est installe sur l'ordinateur.
	// Voir : http://www.boost.org/doc/libs/1_43_0/libs/smart_ptr/shared_ptr.htm

	// Dans ce cas, le rpertoire racine d'installation de "Boost" doit tre ajout dans la liste des
	// dossiers "include" du compilateur (Menu "Outils" -> "Options..." -> "Projets et solutions" ->
	// "Rpertoire de VC++" -> "Afficher les rpertoires pour" : "Fichiers Include").

	#include <boost/weak_ptr.hpp>	// Utilis dans certains projets drivs
	#include <boost/shared_ptr.hpp>

	using namespace boost;

#else					// #if (choix de la source pour "shared_ptr")

	// Dans tous les autres cas, on utilise la version "light" fournie ici.

	// Ce qui suit est une extraction simplifie des templates pour "shared_ptr" et les
	// templates associs.

	// Ceci vite la ncessit d'inclure dans le projet les quelques 48Mo de code source et
	// 7300 fichiers que reprsente la bibliothque "boost" au complet, alors que pour le
	// moment, on n'en utilise que le template "shared_ptr".

	// Il s'agit essentiellement de la concatnation, expurge de quelques lments
	// inutiles dans notre cas, des fichiers "boost" suivants :
	//		- sp_counted_base_w32.hpp
	//		- sp_counted_impl.hpp
	//		- shared_count.hpp
	//		- shared_ptr.hpp
	// Les fichiers originaux incluaient encore d'autres fichiers. Des substitutions ont
	// t ralises dans le code pour viter cette ncessit.


#define BOOST_NO_EXCEPTIONS		// Pas d'exceptions pour le moment (rien n'est prvu pour
								// les recevoir ... mais il faudrait y songer)

#ifdef BOOST_NO_EXCEPTIONS
/// Juste pour satisfaire les appels de ce type dans le code :
inline void throw_exception( std::exception const & e ) // user defined
{
	// rien
}
#endif

// ################################################################################################

// Regroupement des en-ttes ncessaires :
#include <functional>           // std::less
#include <new>                  // std::bad_alloc
#include <algorithm>            // for std::swap
#include <typeinfo>             // for std::bad_cast
#include <memory>               // std::allocator
#include <cstddef>              // std::size_t
#include <iosfwd>               // for std::basic_ostream
#include <ostream>

// ################################################################################################
// begin of "sp_counted_base_w32.hpp" excerpt

//
//  detail/sp_counted_base_w32.hpp
//
//  Copyright (c) 2001, 2002, 2003 Peter Dimov and Multi Media Ltd.
//  Copyright 2004-2005 Peter Dimov
//
//  Distributed under the Boost Software License, Version 1.0. (See
//  accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
//
//  Lock-free algorithm by Alexander Terekhov
//
//  Thanks to Ben Hitchings for the #weak + (#shared != 0)
//  formulation
//

class sp_counted_base
{
private:

	sp_counted_base( sp_counted_base const & );
	sp_counted_base & operator= ( sp_counted_base const & );

	long use_count_;        // #shared
	long weak_count_;       // #weak + (#shared != 0)

public:

	sp_counted_base(): use_count_( 1 ), weak_count_( 1 )
	{
	}

	virtual ~sp_counted_base() // nothrow
	{
	}

	// dispose() is called when use_count_ drops to zero, to release
	// the resources managed by *this.

	virtual void dispose() = 0; // nothrow

	// destroy() is called when weak_count_ drops to zero.

	virtual void destroy() // nothrow
	{
		delete this;
	}

	virtual void * get_deleter( /*std::*/type_info const & ti ) = 0;

	void add_ref_copy()
	{
		InterlockedIncrement( &use_count_ );
	}

	bool add_ref_lock() // true on success
	{
		for( ;; )
		{
			long tmp = static_cast< long const volatile& >( use_count_ );
			if( tmp == 0 ) return false;

			if( InterlockedCompareExchange( &use_count_, tmp + 1, tmp ) == tmp ) return true;
		}
	}

	void release() // nothrow
	{
		if( InterlockedDecrement( &use_count_ ) == 0 )
		{
			dispose();
			weak_release();
		}
	}

	void weak_add_ref() // nothrow
	{
		InterlockedIncrement( &weak_count_ );
	}

	void weak_release() // nothrow
	{
		if( InterlockedDecrement( &weak_count_ ) == 0 )
		{
			destroy();
		}
	}

	long use_count() const // nothrow
	{
		return static_cast<long const volatile &>( use_count_ );
	}
};

// end of "sp_counted_base_w32.hpp" excerpt
// ################################################################################################
// begin of "sp_counted_impl.hpp" excerpt

//
//  detail/sp_counted_impl.hpp
//
//  Copyright (c) 2001, 2002, 2003 Peter Dimov and Multi Media Ltd.
//  Copyright 2004-2005 Peter Dimov
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

template<class X> class sp_counted_impl_p: public sp_counted_base
{
private:

	X * px_;

	sp_counted_impl_p( sp_counted_impl_p const & );
	sp_counted_impl_p & operator= ( sp_counted_impl_p const & );

	typedef sp_counted_impl_p<X> this_type;

public:

	explicit sp_counted_impl_p( X * px ): px_( px )
	{
	}

	virtual void dispose() // nothrow
	{
		delete px_;
	}

	virtual void * get_deleter( /*std::*/type_info const & )
	{
		return 0;
	}

	void * operator new( std::size_t )
	{
		return std::allocator<this_type>().allocate( 1, static_cast<this_type *>(0) );
	}

	void operator delete( void * p )
	{
		std::allocator<this_type>().deallocate( static_cast<this_type *>(p), 1 );
	}
};

template<class P, class D> class sp_counted_impl_pd: public sp_counted_base
{
private:

	P ptr; // copy constructor must not throw
	D del; // copy constructor must not throw

	sp_counted_impl_pd( sp_counted_impl_pd const & );
	sp_counted_impl_pd & operator= ( sp_counted_impl_pd const & );

	typedef sp_counted_impl_pd<P, D> this_type;

public:

	// pre: d(p) must not throw

	sp_counted_impl_pd( P p, D d ): ptr(p), del(d)
	{
	}

	virtual void dispose() // nothrow
	{
		del( ptr );
	}

	virtual void * get_deleter( /*std::*/type_info const & ti )
	{
		return ti == typeid(D)? &reinterpret_cast<char&>( del ): 0;
	}

	void * operator new( std::size_t )
	{
		return std::allocator<this_type>().allocate( 1, static_cast<this_type *>(0) );
	}

	void operator delete( void * p )
	{
		std::allocator<this_type>().deallocate( static_cast<this_type *>(p), 1 );
	}
};

template<class P, class D, class A> class sp_counted_impl_pda: public sp_counted_base
{
private:

	P p_; // copy constructor must not throw
	D d_; // copy constructor must not throw
	A a_; // copy constructor must not throw

	sp_counted_impl_pda( sp_counted_impl_pda const & );
	sp_counted_impl_pda & operator= ( sp_counted_impl_pda const & );

	typedef sp_counted_impl_pda<P, D, A> this_type;

public:

	// pre: d( p ) must not throw

	sp_counted_impl_pda( P p, D d, A a ): p_( p ), d_( d ), a_( a )
	{
	}

	virtual void dispose() // nothrow
	{
		d_( p_ );
	}

	virtual void destroy() // nothrow
	{
		typedef typename A::template rebind< this_type >::other A2;

		A2 a2( a_ );

		this->~this_type();
		a2.deallocate( this, 1 );
	}

	virtual void * get_deleter( /*std::*/type_info const & ti )
	{
		return ti == typeid( D )? &reinterpret_cast<char&>( d_ ): 0;
	}
};
// end of "sp_counted_impl.hpp" excerpt
// ################################################################################################
// begin of "shared_count.hpp" excerpt

//
//  detail/shared_count.hpp
//
//  Copyright (c) 2001, 2002, 2003 Peter Dimov and Multi Media Ltd.
//  Copyright 2004-2005 Peter Dimov
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

struct sp_nothrow_tag {};

class weak_count;

class shared_count
{
private:

	sp_counted_base * pi_;

	friend class weak_count;

public:

	shared_count(): pi_(0) // nothrow
	{
	}

	template<class Y> explicit shared_count( Y * p ): pi_( 0 )
	{
#ifndef BOOST_NO_EXCEPTIONS

		try
		{
			pi_ = new sp_counted_impl_p<Y>( p );
		}
		catch(...)
		{
			delete p;
			throw;
		}

#else

		pi_ = new sp_counted_impl_p<Y>( p );

		if( pi_ == 0 )
		{
			delete p;
			throw_exception( std::bad_alloc() );
		}

#endif
	}

	template<class P, class D> shared_count( P p, D d ): pi_(0)
	{
#ifndef BOOST_NO_EXCEPTIONS

		try
		{
			pi_ = new sp_counted_impl_pd<P, D>(p, d);
		}
		catch(...)
		{
			d(p); // delete p
			throw;
		}

#else

		pi_ = new sp_counted_impl_pd<P, D>(p, d);

		if(pi_ == 0)
		{
			d(p); // delete p
			throw_exception(std::bad_alloc());
		}

#endif
	}

	template<class P, class D, class A> shared_count( P p, D d, A a ): pi_( 0 )
	{
		typedef sp_counted_impl_pda<P, D, A> impl_type;
		typedef typename A::template rebind< impl_type >::other A2;

		A2 a2( a );

#ifndef BOOST_NO_EXCEPTIONS

		try
		{
			pi_ = a2.allocate( 1, static_cast< impl_type* >( 0 ) );
			new( static_cast< void* >( pi_ ) ) impl_type( p, d, a );
		}
		catch(...)
		{
			d( p );

			if( pi_ != 0 )
			{
				a2.deallocate( static_cast< impl_type* >( pi_ ), 1 );
			}

			throw;
		}

#else

		pi_ = a2.allocate( 1, static_cast< impl_type* >( 0 ) );

		if( pi_ != 0 )
		{
			new( static_cast< void* >( pi_ ) ) impl_type( p, d, a );
		}
		else
		{
			d( p );
			throw_exception( std::bad_alloc() );
		}

#endif
	}

	// auto_ptr<Y> is special cased to provide the strong guarantee

	template<class Y>
	explicit shared_count( std::auto_ptr<Y> & r ): pi_( new sp_counted_impl_p<Y>( r.get() ) )
	{
#ifdef BOOST_NO_EXCEPTIONS

		if( pi_ == 0 )
		{
			isMultiplexProgthrow_exception(std::bad_alloc());
		}

#endif

		r.release();
	}

	~shared_count() // nothrow
	{
		if( pi_ != 0 ) pi_->release();
	}

	shared_count(shared_count const & r): pi_(r.pi_) // nothrow
	{
		if( pi_ != 0 ) pi_->add_ref_copy();
	}

	explicit shared_count(weak_count const & r); // throws bad_weak_ptr when r.use_count() == 0
	shared_count( weak_count const & r, sp_nothrow_tag ); // constructs an empty *this when r.use_count() == 0

	shared_count & operator= (shared_count const & r) // nothrow
	{
		sp_counted_base * tmp = r.pi_;

		if( tmp != pi_ )
		{
			if( tmp != 0 ) tmp->add_ref_copy();
			if( pi_ != 0 ) pi_->release();
			pi_ = tmp;
		}

		return *this;
	}

	void swap(shared_count & r) // nothrow
	{
		sp_counted_base * tmp = r.pi_;
		r.pi_ = pi_;
		pi_ = tmp;
	}

	long use_count() const // nothrow
	{
		return pi_ != 0? pi_->use_count(): 0;
	}

	bool unique() const // nothrow
	{
		return use_count() == 1;
	}

	bool empty() const // nothrow
	{
		return pi_ == 0;
	}

	friend inline bool operator==(shared_count const & a, shared_count const & b)
	{
		return a.pi_ == b.pi_;
	}

	friend inline bool operator<(shared_count const & a, shared_count const & b)
	{
		return std::less<sp_counted_base *>()( a.pi_, b.pi_ );
	}

	void * get_deleter( /*std::*/type_info const & ti ) const
	{
		return pi_? pi_->get_deleter( ti ): 0;
	}
};


class weak_count
{
private:

	sp_counted_base * pi_;

	friend class shared_count;

public:

	weak_count(): pi_(0) // nothrow
	{
	}

	weak_count(shared_count const & r): pi_(r.pi_) // nothrow
	{
		if(pi_ != 0) pi_->weak_add_ref();
	}

	weak_count(weak_count const & r): pi_(r.pi_) // nothrow
	{
		if(pi_ != 0) pi_->weak_add_ref();
	}

	~weak_count() // nothrow
	{
		if(pi_ != 0) pi_->weak_release();
	}

	weak_count & operator= (shared_count const & r) // nothrow
	{
		sp_counted_base * tmp = r.pi_;

		if( tmp != pi_ )
		{
			if(tmp != 0) tmp->weak_add_ref();
			if(pi_ != 0) pi_->weak_release();
			pi_ = tmp;
		}

		return *this;
	}

	weak_count & operator= (weak_count const & r) // nothrow
	{
		sp_counted_base * tmp = r.pi_;

		if( tmp != pi_ )
		{
			if(tmp != 0) tmp->weak_add_ref();
			if(pi_ != 0) pi_->weak_release();
			pi_ = tmp;
		}

		return *this;
	}

	void swap(weak_count & r) // nothrow
	{
		sp_counted_base * tmp = r.pi_;
		r.pi_ = pi_;
		pi_ = tmp;
	}

	long use_count() const // nothrow
	{
		return pi_ != 0? pi_->use_count(): 0;
	}

	bool empty() const // nothrow
	{
		return pi_ == 0;
	}

	friend inline bool operator==(weak_count const & a, weak_count const & b)
	{
		return a.pi_ == b.pi_;
	}

	friend inline bool operator<(weak_count const & a, weak_count const & b)
	{
		return std::less<sp_counted_base *>()(a.pi_, b.pi_);
	}
};

inline shared_count::shared_count( weak_count const & r ): pi_( r.pi_ )
{
	if( pi_ == 0 || !pi_->add_ref_lock() )
	{
	    /*throw_exception( bad_weak_ptr() )*/;
	}
}

inline shared_count::shared_count( weak_count const & r, sp_nothrow_tag ): pi_( r.pi_ )
{
	if( pi_ != 0 && !pi_->add_ref_lock() )
	{
		pi_ = 0;
	}
}

// end of  "shared_count.hpp" excerpt
// ################################################################################################
// begin of "shared_ptr.hpp" excerpt

//
//  shared_ptr.hpp
//
//  (C) Copyright Greg Colvin and Beman Dawes 1998, 1999.
//  Copyright (c) 2001-2008 Peter Dimov
//
//  Distributed under the Boost Software License, Version 1.0. (See
//  accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
//  See http://www.boost.org/libs/smart_ptr/shared_ptr.htm for documentation.
//

template<class T> class shared_ptr;
template<class T> class weak_ptr;
template<class T> class enable_shared_from_this;
template<class T> class enable_shared_from_this2;

struct static_cast_tag {};
struct const_cast_tag {};
struct dynamic_cast_tag {};
struct polymorphic_cast_tag {};

template<class T> struct shared_ptr_traits
{
	typedef T & reference;
};

template<> struct shared_ptr_traits<void>
{
	typedef void reference;
};

template<> struct shared_ptr_traits<void const>
{
	typedef void reference;
};

template<> struct shared_ptr_traits<void volatile>
{
	typedef void reference;
};

template<> struct shared_ptr_traits<void const volatile>
{
	typedef void reference;
};

// enable_shared_from_this support

template< class X, class Y, class T > inline void sp_enable_shared_from_this( shared_ptr<X> const * ppx, Y const * py, enable_shared_from_this< T > const * pe )
{
	if( pe != 0 )
	{
		pe->_internal_accept_owner( ppx, const_cast< Y* >( py ) );
	}
}

template< class X, class Y, class T > inline void sp_enable_shared_from_this( shared_ptr<X> * ppx, Y const * py, enable_shared_from_this2< T > const * pe )
{
	if( pe != 0 )
	{
		pe->_internal_accept_owner( ppx, const_cast< Y* >( py ) );
	}
}

inline void sp_enable_shared_from_this( ... )
{
}

// rvalue auto_ptr support based on a technique by Dave Abrahams

template< class T, class R > struct sp_enable_if_auto_ptr
{
};

template< class T, class R > struct sp_enable_if_auto_ptr< std::auto_ptr< T >, R >
{
	typedef R type;
}; 

//
//  shared_ptr
//
//  An enhanced relative of scoped_ptr with reference counted copy semantics.
//  The object pointed to is deleted when the last shared_ptr pointing to it
//  is destroyed or reset.
//

template<class T> class shared_ptr
{
private:

	// Borland 5.5.1 specific workaround
	typedef shared_ptr<T> this_type;

public:

	typedef T element_type;
	typedef T value_type;
	typedef T * pointer;
	typedef typename shared_ptr_traits<T>::reference reference;

	shared_ptr(): px(0), pn() // never throws in 1.30+
	{
	}

	template<class Y>
	explicit shared_ptr( Y * p ): px( p ), pn( p ) // Y must be complete
	{
		sp_enable_shared_from_this( this, p, p );
	}

	//
	// Requirements: D's copy constructor must not throw
	//
	// shared_ptr will release p by calling d(p)
	//

	template<class Y, class D> shared_ptr(Y * p, D d): px(p), pn(p, d)
	{
		sp_enable_shared_from_this( this, p, p );
	}

	// As above, but with allocator. A's copy constructor shall not throw.

	template<class Y, class D, class A> shared_ptr( Y * p, D d, A a ): px( p ), pn( p, d, a )
	{
		sp_enable_shared_from_this( this, p, p );
	}

//  generated copy constructor, destructor are fine

	template<class Y>
	explicit shared_ptr(weak_ptr<Y> const & r): pn(r.pn) // may throw
	{
		// it is now safe to copy r.px, as pn(r.pn) did not throw
		px = r.px;
	}

	template<class Y>
	shared_ptr( weak_ptr<Y> const & r, sp_nothrow_tag ): px( 0 ), pn( r.pn, sp_nothrow_tag() ) // never throws
	{
	    if( !pn.empty() )
	    {
	        px = r.px;
	    }
	}

	template<class Y>
	shared_ptr( shared_ptr<Y> const & r )
	: px( r.px ), pn( r.pn ) // never throws
	{
	}

	// aliasing
	template< class Y >
	shared_ptr( shared_ptr<Y> const & r, T * p ): px( p ), pn( r.pn ) // never throws
	{
	}

	template<class Y>
	shared_ptr(shared_ptr<Y> const & r, static_cast_tag): px(static_cast<element_type *>(r.px)), pn(r.pn)
	{
	}

	template<class Y>
	shared_ptr(shared_ptr<Y> const & r, const_cast_tag): px(const_cast<element_type *>(r.px)), pn(r.pn)
	{
	}

	template<class Y>
	shared_ptr(shared_ptr<Y> const & r, dynamic_cast_tag): px(dynamic_cast<element_type *>(r.px)), pn(r.pn)
	{
		if(px == 0) // need to allocate new counter -- the cast failed
		{
			pn = shared_count();
		}
	}

	template<class Y>
	shared_ptr(shared_ptr<Y> const & r, polymorphic_cast_tag): px(dynamic_cast<element_type *>(r.px)), pn(r.pn)
	{
		if(px == 0)
		{
			throw_exception(std::bad_cast());
		}
	}
	template<class Y>
	explicit shared_ptr(std::auto_ptr<Y> & r): px(r.get()), pn()
	{
		Y * tmp = r.get();
		pn = shared_count(r);
		sp_enable_shared_from_this( this, tmp, tmp );
	}
	template<class Ap>
	explicit shared_ptr( Ap r, typename sp_enable_if_auto_ptr<Ap, int>::type = 0 ): px( r.get() ), pn()
	{
		typename Ap::element_type * tmp = r.get();
		pn = shared_count( r );
		sp_enable_shared_from_this( this, tmp, tmp );
	}

	// assignment

	shared_ptr & operator=( shared_ptr const & r ) // never throws
	{
		this_type(r).swap(*this);
		return *this;
	}

#if !defined(BOOST_MSVC) || (BOOST_MSVC >= 1400)

	template<class Y>
	shared_ptr & operator=(shared_ptr<Y> const & r) // never throws
	{
		this_type(r).swap(*this);
		return *this;
	}

#endif

	template<class Y>
	shared_ptr & operator=( std::auto_ptr<Y> & r )
	{
		this_type(r).swap(*this);
		return *this;
	}

	template<class Ap>
	typename sp_enable_if_auto_ptr< Ap, shared_ptr & >::type operator=( Ap r )
	{
		this_type( r ).swap( *this );
		return *this;
	}

	void reset() // never throws in 1.30+
	{
		this_type().swap(*this);
	}

	template<class Y> void reset(Y * p) // Y must be complete
	{
		ASSERT(p == 0 || p != px); // catch self-reset errors
		this_type(p).swap(*this);
	}

	template<class Y, class D> void reset( Y * p, D d )
	{
		this_type( p, d ).swap( *this );
	}

	template<class Y, class D, class A> void reset( Y * p, D d, A a )
	{
		this_type( p, d, a ).swap( *this );
	}

	template<class Y> void reset( shared_ptr<Y> const & r, T * p )
	{
		this_type( r, p ).swap( *this );
	}

	reference operator* () const // never throws
	{
		ASSERT(px != 0);
		return *px;
	}

	T * operator-> () const // never throws
	{
		ASSERT(px != 0);
		return px;
	}

	T * get() const // never throws
	{
		return px;
	}

	typedef T * this_type::*unspecified_bool_type;

	operator unspecified_bool_type() const // never throws
	{
		return px == 0? 0: &this_type::px;
	}

	// operator! is redundant, but some compilers need it
	bool operator! () const // never throws
	{
		return px == 0;
	}

	bool unique() const // never throws
	{
		return pn.unique();
	}

	long use_count() const // never throws
	{
		return pn.use_count();
	}

	void swap(shared_ptr<T> & other) // never throws
	{
		std::swap(px, other.px);
		pn.swap(other.pn);
	}

	template<class Y> bool _internal_less(shared_ptr<Y> const & rhs) const
	{
		return pn < rhs.pn;
	}

	void * _internal_get_deleter( /*std::*/type_info const & ti ) const
	{
		return pn.get_deleter( ti );
	}

	bool _internal_equiv( shared_ptr const & r ) const
	{
		return px == r.px && pn == r.pn;
	}

// Tasteless as this may seem, making all members public allows member templates
// to work in the absence of member template friends. (Matthew Langston)

private:

	template<class Y> friend class shared_ptr;
	template<class Y> friend class weak_ptr;

	T * px;             // contained pointer
	shared_count pn;    // reference counter

};	// shared_ptr

template<class T, class U> inline bool operator==(shared_ptr<T> const & a, shared_ptr<U> const & b)
{
	return a.get() == b.get();
}

template<class T, class U> inline bool operator!=(shared_ptr<T> const & a, shared_ptr<U> const & b)
{
	return a.get() != b.get();
}

template<class T, class U> inline bool operator<(shared_ptr<T> const & a, shared_ptr<U> const & b)
{
	return a._internal_less(b);
}

template<class T> inline void swap(shared_ptr<T> & a, shared_ptr<T> & b)
{
	a.swap(b);
}

template<class T, class U> shared_ptr<T> static_pointer_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, static_cast_tag());
}

template<class T, class U> shared_ptr<T> const_pointer_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, const_cast_tag());
}

template<class T, class U> shared_ptr<T> dynamic_pointer_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, dynamic_cast_tag());
}

// shared_*_cast names are deprecated. Use *_pointer_cast instead.

template<class T, class U> shared_ptr<T> shared_static_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, static_cast_tag());
}

template<class T, class U> shared_ptr<T> shared_dynamic_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, dynamic_cast_tag());
}

template<class T, class U> shared_ptr<T> shared_polymorphic_cast(shared_ptr<U> const & r)
{
	return shared_ptr<T>(r, polymorphic_cast_tag());
}

template<class T, class U> shared_ptr<T> shared_polymorphic_downcast(shared_ptr<U> const & r)
{
	BOOST_ASSERT(dynamic_cast<T *>(r.get()) == r.get());
	return shared_static_cast<T>(r);
}

// get_pointer() enables mem_fn to recognize shared_ptr

template<class T> inline T * get_pointer(shared_ptr<T> const & p)
{
	return p.get();
}

// operator<<

template<class E, class T, class Y> std::basic_ostream<E, T> & operator<< (std::basic_ostream<E, T> & os, shared_ptr<Y> const & p)
{
	os << p.get();
	return os;
}

template<class D, class T> D * get_deleter(shared_ptr<T> const & p)
{
	return static_cast<D *>(p._internal_get_deleter(typeid(D)));
}
// end of "shared_ptr.hpp" excerpt
// ################################################################################################

#endif					// #if (choix de la source pour "shared_ptr")
