/*
 * $Id: channels.cpp 399 2020-10-22 17:12:06Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation de fonctions lies  la gestion des chanes TV
 *
 *	Ce fichier implmente les lments ncessaires au stockage et  la gestion
 *	des informations lies aux chanes TV identifies par l'application.
 **/

#include "stdafx.h"
#include "channels.h"
#include "record.h"
#include "crc32.h"

/// Tableau des chanes
Chaines		Canaux;

int			ixChaineCourante	= -1;	//!< Index de la chane couramment utilise
int			ixChaineAvance		= -1;	/*!< Index de la chane pendant les transitions de zapping
											 (normalement identique  \p ixChaineCourante, sauf entre la
											 slection d'une chane et l'application de cette slection) */
int			ixDerniereChaine	= -1;	//!< Mmorisation dernire chane regarde (avant chane courante)
int			ixSonCourant		= 0;	//!< Index de la piste sonore couramment utilise
bool		suspendu			= true;	//!< \p true si vido et audio suspendus ( l'initialisation,
										//!< ou si minimis + \p suspend_minimized)

// Taille des images des chanes (largeur et longueur)
#define TAILLE_IMAGE_CHAINE 16

static HBITMAP cherche_icone(LPCTSTR nom, bool audio)
{
	SZPathName	fileName;
	HBITMAP		hImage	= NULL;
	tstring		strChemin;

	// Concatnation chemin et chane
	_stprintf_s(fileName, TEXT("%s.bmp"), nom);

	if (FindAppFile(TEXT("Icones"), fileName, strChemin)!=pd_notfound) {
		// Chargement de l'icne, paramtr de telle sorte (grce  LR_LOADTRANSPARENT + LR_LOADMAP3DCOLORS)
		// que le premier pixel de l'image serve de couleur de rfrence pour une pseudo-transparence de celle-ci,
		//  savoir qu'elle sera remplace par la couleur de fond des menus.
		hImage =	static_cast<HBITMAP>(LoadImage(NULL, strChemin.c_str(), IMAGE_BITMAP,
						TAILLE_IMAGE_CHAINE, TAILLE_IMAGE_CHAINE, LR_LOADFROMFILE | LR_LOADTRANSPARENT | LR_LOADMAP3DCOLORS));
		myprintf(CMyP FMT_ERRX("LoadImage %s"), hImage==NULL, strChemin.c_str(), GetLastError());
	}

	if (hImage == NULL) {
		LPCTSTR intres = audio ? MAKEINTRESOURCE(IDB_RADIO_GENERIC) : MAKEINTRESOURCE(IDB_TV_GENERIC);

		// En l'absence d'icne pour cette chane, utiliser une icne gnrique en niveaux de gris
		hImage = static_cast<HBITMAP>(LoadImage(hAppInstance, intres, IMAGE_BITMAP,
			TAILLE_IMAGE_CHAINE, TAILLE_IMAGE_CHAINE, LR_LOADTRANSPARENT | LR_LOADMAP3DCOLORS));
		if (hImage == NULL) {
			myprintf(FMT_ERRX("LoadBitmap gnrique"), GetLastError());
			// Pas trouv : on met INVALID_HANDLE_VALUE pour se rappeler de ne plus le
			// rechercher avant le prochain lancement du programme.
			hImage = (HBITMAP)INVALID_HANDLE_VALUE;
		}
	}

	return hImage;
}

/// Constructeur
Chaine::Chaine() :
	khz(-1),
	nNumeroNit(0),
	nNumeroAuto(0),
	pIconCache(new IconHCache)
{
	nom[0] = 0;
	groupe[0] = 0;
}

/// Destructeur
Chaine::~Chaine()
{
}

#ifdef _XMLUTILS_H_INCLUDED_

/// Noms des tags XML utiliss pour la sauvegarde d'une chane
LPCSTR Chaine::pszTagStrList =
	"Nom\0"
	"Canal\0"
	"Groupe\0"
	"Frequence\0"
	"ONID\0"
	"TSID\0"
	"SID\0"
	"Numero_initial\0"
	"Numero_auto\0"
	"Numero\0"
	"Etat\0"
	"PMT\0"
	"PCR\0"
  #if STREAMS_SAVE_ALL
	"Video\0"
	"Video_mpeg4\0"
	"Video_type\0"
  #endif
	;

#if USE_PMTFILTER
void CALLBACK Chaine::PmtMonitor(const ChainePMT & sExpectedPmt, const ChainePMT & sFoundPmt, DWORD_PTR dwUserData)
{
	int ixChaine = Canaux.trouve_par_cle((UINT32)dwUserData);

	if (!ixChaine_ok(ixChaine))
		return;

	dynamic_cast<ChainePMT &>(Canaux[ixChaine]) = sFoundPmt;	// Seulement la base "ChainePMT"

	// Notifier du changement de PMT (demande de rezapper
	// sur cette chane, et de sauvegarder les modifications)
	PostMessage(hMainWnd, WM_APP_PMT_UPDATED, 0, 0);
}
#endif

/**
 * Chargement de la dfinition de la chane  partir d'un noeud TinyXML construit
 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
 * \return		\p true ou \p false selon succs ou chec
 **/
bool Chaine::SetFromXMLElement(const TiXmlElement * pXmlElm)
{
	for (ITERATE_XML_ELEMENTS(pXmlElm, pXmlChild)) {
		LPCSTR	pszTag = pXmlChild->Value();
		LPCSTR	pszTxt = pXmlChild->GetText();

		if (pszTxt == NULL)
			pszTxt = "0";

		// Conversion en numrique pour les champs qui le requirent.
		long	nVal = atol(pszTxt);

		switch (StringToEnum(pszTagStrList, pszTag)) {

		case tlc_Nom:				strcpy_T(nom, ConvertUTF82Char(pszTxt).c_str());	break;
		case tlc_Canal:				canal_no = (UINT16)nVal;							break;
		case tlc_Groupe:			strcpy_T(groupe, ConvertUTF82Char(pszTxt).c_str());	break;
		case tlc_Frequence:			khz = nVal;											break;
		case tlc_ONID:				ONID = (UINT16)nVal;								break;
		case tlc_TSID:				TSID = (UINT16)nVal;								break;
		case tlc_SID:				SID = (UINT16)nVal;									break;
		case tlc_Numero_initial:	nNumeroNit = (UINT16)nVal;							break;
		case tlc_Numero_auto:		nNumeroAuto = (UINT16)nVal;							break;
		case tlc_Numero:			nNumeroChaine = (UINT16)nVal;						break;
		case tlc_Etat:				etat = (EtatChaine)str2dword(aChStateTable,
										ConvertUTF82Char(pszTxt).c_str(), ec_preferee); break;
		case tlc_PMT:				pmt_pid = (UINT16)nVal;								break;
		case tlc_PCR:				pcr_pid = (UINT16)nVal;								break;

	  #if STREAMS_SAVE_ALL
		case tlc_Video:
			video_pid = (UINT16)nVal;
			if (video_type == vst_Unknown)
				video_type = vst_MPEG2;	// pour compatibilit versions antrieures  SVN 175
			break;

		case tlc_Video_mpeg4:
			// Tag pour compatibilit versions antrieures  SVN 175
			video_pid = (UINT16)nVal;
			if (video_type == vst_Unknown)
				video_type = vst_H264;
			break;

		case tlc_Video_type:
			video_type = (VideoStreamType)str2dword(aVideoTypeTable,
				ConvertUTF82Char(pszTxt).c_str(), vst_Unknown);
	  #endif	// #if STREAMS_SAVE_ALL
		}
	}

  #if STREAMS_SAVE_ALL
	// Rcupre les pistes son
	SetStreamsFromXML(pXmlElm, false);
	// Rcupre les autres informations
	SetStreamsFromXML(pXmlElm, true);
  #endif
	return true;
}

/**
 * Construction d'un noeud TinyXML reprsentant la dfinition de la chane
 * \return	Noeud XML construit
 **/
TiXmlElement * Chaine::GetAsXMLElement() const
{
	TiXmlElement *	pXmlChan(new TiXmlElement("Chaine"));

	if (pXmlChan) {
		for (int eTag = tlc_Nom; eTag < tlc_END; ++eTag) {
			CHAR szTmp[_countof(nom)] = "";

			switch (eTag) {
			case tlc_Nom:				strcpy_T(szTmp, ConvertChar2UTF8(nom).c_str());		break;
			case tlc_Canal:				_itoa_s(canal_no, szTmp, 10);						break;
			case tlc_Groupe:			strcpy_T(szTmp, ConvertChar2UTF8(groupe).c_str());	break;
			case tlc_Frequence:			_itoa_s(khz, szTmp, 10);							break;
			case tlc_ONID:				_itoa_s(ONID, szTmp, 10);							break;
			case tlc_TSID:				_itoa_s(TSID, szTmp, 10);							break;
			case tlc_SID:				_itoa_s(SID, szTmp, 10);							break;
			case tlc_Numero_initial:	_itoa_s(nNumeroNit, szTmp, 10);						break;
			case tlc_Numero_auto:		_itoa_s(nNumeroAuto, szTmp, 10);					break;
			case tlc_Numero:			_itoa_s(nNumeroChaine, szTmp, 10);					break;
			case tlc_Etat:				strcpy_T(szTmp, ConvertChar2UTF8(dword2str(aChStateTable, etat)).c_str());
																							break;
			case tlc_PMT:				_itoa_s(pmt_pid, szTmp, 10);						break;
			case tlc_PCR:				_itoa_s(pcr_pid, szTmp, 10);						break;
	  #if STREAMS_SAVE_ALL
			case tlc_Video:				_itoa_s(video_pid, szTmp, 10);						break;
			case tlc_Video_mpeg4:		continue;
			case tlc_Video_type:		strcpy_T(szTmp, ConvertChar2UTF8(dword2str(aVideoTypeTable, video_type)).c_str());
	  #endif
			}

			SetXmlContent(pXmlChan, EnumToString(eTag, pszTagStrList), szTmp);
		}
	}

  #if STREAMS_SAVE_ALL
	// Mmorise toutes les pistes son
	GetStreamsAsXML(pXmlChan, false);
	// Mmorise toutes les autres informations
	GetStreamsAsXML(pXmlChan, true);
  #endif
	return pXmlChan;
}

#if STREAMS_SAVE_ALL

/**
 * Chargement de la dfinition d'un groupe de pistes  partir d'un noeud TinyXML construit
 * \param[in]	pXmlElm	Noeud XML parent de celui contenant les informations recherches
 * \param[in]	bOther	\p false pour pistes "son", \p true pour pistes "autre"
 * \return		\p true ou \p false selon succs ou chec
 **/
bool Chaine::SetStreamsFromXML(const TiXmlElement * pXmlElm, bool bOther)
{
	LPCSTR					pszTagStr	= getTagBase(bOther);
	const TiXmlElement *	pChild		= NULL;
	int						nCount		= (int)atol(GetXmlContentF(pXmlElm, &pChild, "Nb_%s", pszTagStr));

	for (int i=0; i < nCount; i++) {
		une_piste			sPiste;

		sPiste.pid	= (UINT16)atol(GetXmlContentF(pXmlElm, &pChild, "%s%i", pszTagStr, i));
		sPiste.type	= (TrackStreamType)atol(GetXmlContentF(pXmlElm, &pChild, "%s%i_Type", pszTagStr, i));
		strcpy_T(sPiste.lang, GetXmlContentF(pXmlElm, &pChild, "%s%i_Lang", pszTagStr, i));
		ajoute(sPiste);
	}
	return true;
}

/**
 * Ajout des lments de ce groupe de pistes  un noeud TinyXML
 * \param[in]	pXmlElm	Noeud XML auquel ajouter les informations
 * \param[in]	bOther	\p false pour pistes "son", \p true pour pistes "autre"
 **/
void Chaine::GetStreamsAsXML(TiXmlElement * pXmlElm, bool bOther) const
{
	LPCSTR	pszTagStr	= getTagBase(bOther);
	int		nMax		= getCount(bOther);
	int		nOfs		= getFirst(bOther);

	CHAR szTmp[16];

	_itoa_s(nMax, szTmp, 10);
	SetXmlContentF(pXmlElm, "Nb_%s", szTmp, pszTagStr);

	for (int j=0; j < nMax; j++) {
		const une_piste &	sPiste = tbl[j+nOfs];
		TiXmlElement *		pChild;

		_itoa_s(sPiste.pid, szTmp, 10);
		pChild = SetXmlContentF(pXmlElm, "%s%i", szTmp, pszTagStr, j);
		_itoa_s(sPiste.type, szTmp, 10);
		pChild = SetXmlContentF(pXmlElm, "%s%i_Type", szTmp, pszTagStr, j);
		if (*sPiste.lang)
			SetXmlContentF(pXmlElm, "%s%i_Lang", sPiste.lang, pszTagStr, j);
	}
}

#endif	//  #if STREAMS_SAVE_ALL

#endif	// #ifdef _XMLUTILS_H_INCLUDED_

/// Obtenir l'icne associe  la chane, au format bitmap
HBITMAP Chaine::GetIconAsBmp() const
{
	ASSERT(pIconCache.get() != NULL);

	HBITMAP & hImage = pIconCache->hImage;

	if (hImage==INVALID_HANDLE_VALUE)
		return NULL;

	if (hImage==NULL) {
		/**
		 * Charge les bitmaps des chanes
		 **/
		NomProtege<TCHAR> nom_canal(nom, '_');

		hImage =  cherche_icone(nom_canal.nom, isAudio());
	}

	return hImage;
}

/// Obtenir l'icne associe  la chane, au format ICO
HICON Chaine::GetIconAsIco() const
{
	HBITMAP hImage = GetIconAsBmp();

	if (hImage != NULL) {
		AutoDC			cDC(hMainWnd);
		BITMAPINFO		sBmpInf = {{sizeof(sBmpInf.bmiHeader)}};

		if (GetDIBits(cDC, hImage, 0, 0, NULL, &sBmpInf, DIB_RGB_COLORS)) {
			ICONINFO	sIconInfo = {
				TRUE,									// fIcon
				0, 0,									// xHotspot, yHotspot
				::CreateCompatibleBitmap(cDC,			// hbmMask
					sBmpInf.bmiHeader.biWidth, sBmpInf.bmiHeader.biHeight),
				hImage									// hbmColor
			};

			if (sIconInfo.hbmMask) {
				pIconCache->hIcon = ::CreateIconIndirect(&sIconInfo);
				::DeleteObject(sIconInfo.hbmMask);
			}
		} else
			myprintf(ERRX(TEXT("chec GetIconAsIco/GetDIBits : %s")), SystemErrorToString().c_str());
	}
	return pIconCache->hIcon;
}

/**
 * \brief Comparaison avec une autre chane pour tri.
 *
 * \retval	Une valeur ngative si cette chane a un rang infrieur  l'autre chane
 * \retval	0 Si cette chane a un rang gal  l'autre chane
 * \retval	Une valeur positive si cette chane a un rang suprieur  l'autre chane
 **/
int Chaine::compare(const Chaine & canal2)	const
{
	if (nNumeroChaine < canal2.nNumeroChaine)
		return -1;
	if (nNumeroChaine > canal2.nNumeroChaine)
		return 1;
	return _stricmp(nom, canal2.nom) > 0;
}

/**
 * Gnration du nom affichable d'une piste associe pour le menu
 * \param[in]	nTrack	Numro de la piste
 * \param[in]	pstr	Pointeur sur le tampon qui recevra le rsultat
 * \param[in]	bufSize	Taille du tampon
 * \param[in]	nNum	Numro de piste  insrer (si >= 0)
 **/
void Chaine::toTrackMenuString(int nTrack, LPTSTR pstr, size_t bufSize, int nNum) const
{
	struct Langue {
		char	code[4];
		LPCSTR	szLang;
	};

	static const Langue Tab_Lang[] =
	{
		{"fra", "Franais"},
		{"fre", "Franais"},
		{"eng", "Anglais"},
		{"ang", "Anglais"},
		{"deu", "Allemand"},
		{"ger", "Allemand"},
		{"spa", "Espagnol"},
		{"ita", "Italien"},
		{"por", "Portugais"},
		{"qaa", "V.O."},
		{"qad", "Audio Descr."}
	};

	static const LPCSTR tblNomsAudio[] =
	{
		"Mpeg",
		"AC3",
		"E-AC3",
		"AC3 HC",	// HC = Home Cinema
		"E-AC3 HC"
	};

	const une_piste &	son_courant		= tbl[nTrack];
	UINT				nLen;

	if (nNum >= 0)
		_stprintf_s(pstr, bufSize, TEXT("Piste %i : "), nNum);
	else
		*pstr = 0;

	for (int j=0; j < _countof(Tab_Lang); j++) {
		const Langue & tlang = Tab_Lang[j];

		if (strcmp(son_courant.lang, tlang.code) == 0) {
			strcat_T(pstr, (UINT)bufSize, tlang.szLang);
			break;
		}
	}

	if (*pstr == 0)
		strcat_T(pstr, (UINT)bufSize, TEXT("Langue non spcifie"));

	strcat_T(pstr, (UINT)bufSize, TEXT(" ("));

	if (son_courant.type < tst_Audio_END) {
		strcat_T(pstr, (UINT)bufSize, tblNomsAudio[son_courant.type]);
	} else {
		nLen = (UINT)_tcslen(pstr);
		_stprintf_s(pstr+nLen, bufSize-nLen, TEXT("Inconnu, code 0x%02x"), son_courant.type);
	}

	strcat_T(pstr, (UINT)bufSize, TEXT(")"));
}

/**
 * Dtermination du mappage de tous les PIDs de la chane ;
 * Si \p reset_son est \p true, alors le canal son est, en mme temps, remis  sa
 * valeur par dfaut, selon prfrences globales de l'utilisateur
 **/
MappingData Chaine::get_MappingData(bool reset_son) const
{
	if (reset_son)
		ixSonCourant = selection_son(audio_pref);

	const une_piste son		= tbl[ixSonCourant];

	return MappingData(pmt_pid, video_type, son.type, video_pid, son.pid);
}

/**
 * Mappage du PID pmt et dbranchement de tous les autres
 **/
HRESULT Chaine::suspend() const
{
	return branche_medias(MappingData(pmt_pid));
}

/**
 * Mappage de tous les PIDs de la chane (la frquence est prsume dj syntonise) ;
 * Si \p reset_son est \p true, alors le canal son est, en mme temps, remis  sa
 * valeur par dfaut, selon prfrences globales de l'utilisateur
 **/
HRESULT Chaine::branche(bool reset_son) const
{
	if (suspendu)
		return S_OK;	// Ne pas brancher si suspendu
	myprintf(TEXT("Chaine::branche SID=%i, TSID=%i, ONID=%i, freq=%i") EOL, SID, TSID, ONID, khz);

	HRESULT hr = branche_medias(get_MappingData(reset_son));

#if USE_PMTFILTER
	if (SUCCEEDED(hr)) {
		CComPtr<IPMTMonitor> pPmtMonitor(pGraph->GetPMTMonitor());

		if (pPmtMonitor)
			hr = pPmtMonitor->StartPmtMonitoring(*this, PmtMonitor, cle());
	}
#endif
	return hr;
}

#ifdef _XMLUTILS_H_INCLUDED_

/**
 * Chargement de l'ensemble des chanes  partir d'un noeud TinyXML construit
 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
 * \return		\p true ou \p false selon succs ou chec
 **/
bool Chaines::SetFromXMLElement(const TiXmlElement * pXmlElm)
{
	if (_stricmp(pXmlElm->Value(), "Chaines")!=0)
		return false;
	for (ITERATE_XML_ELEMENTS_NAMED(pXmlElm, pXmlChild, "Chaine")) {
		Chaine	cChan;

		if (cChan.SetFromXMLElement(pXmlChild))
			push_back(cChan);
	}

	return true;
}

/**
 * Construction d'un noeud TinyXML reprsentant la dfinition de l'ensemble des chanes
 * \return	Noeud XML construit
 **/
TiXmlElement * Chaines::GetAsXMLElement() const
{
	TiXmlElement *	pXmlChLst(new TiXmlElement("Chaines"));

	if (pXmlChLst) {
		for (ITERATE_CONST_VECTOR(*this, Chaine, it)) {
			TiXmlElement *	pXmlChan(it->GetAsXMLElement());

			if (pXmlChan)
				pXmlChLst->LinkEndChild(pXmlChan);
		}
	}

	return pXmlChLst;
}

#endif	// #ifdef _XMLUTILS_H_INCLUDED_

/**
 * Recherche d'une chane en partant de son numro
 * \param[in]	nNumChaine	Numro de la chane, tel que dfini dans la configuration des
 *							chanes et tel qu'affich  l'utilisateur
 * \retval		Index de la chane dans la table #Canaux
 * \retval		-1 si le numro n'a pas t trouv
 **/
int Chaines::trouve_par_no(WORD nNumChaine) const
{
	int nbChaines = (int)size();

	for (int i=0; i<nbChaines; i++) {
		if ((*this)[i].nNumeroChaine == nNumChaine)
			return i;
	}
	return -1;
}

/// \deprecated	(voir dfinition)
int Chaines::trouve_par_sid(WORD sidChaine) const
{
	int nbChaines = (int)size();

	for (int i=0; i<nbChaines; i++) {
		if ((*this)[i].SID == sidChaine)
			return i;
	}
	return -1;
}

/**
 * Recherche d'une chane en partant d'une cl 32 bits
 * \param[in]	cleChaine	Cl 32 bits de la chane, gnre par ChaineIDs::cle
 * \retval		Index de la chane dans la table #Canaux
 * \retval		-1 si le numro n'a pas t trouv
 **/
int Chaines::trouve_par_cle(UINT32 cleChaine) const
{
	if (cleChaine == STREAM_PSEUDO_CLE)
		return STREAM_PSEUDO_INDEX;

	int nbChaines = (int)size();

	for (int i=0; i<nbChaines; i++) {
		if ((*this)[i].cle() == cleChaine)
			return i;
	}
	return -1;
}

/**
 * Recherche d'une restauration de chane en partant d'une cl 32 bits
 * (identique  \p trouve_par_cle, mais retourne la premire chane de la liste
 * si pas trouv, et ne retourne -1 que si aucune chane n'existe)
 * \param[in]	cleChaine	Cl 32 bits de la chane, gnre par ChaineIDs::cle
 * \retval		Index de la chane dans la table #Canaux
 * \retval		-1 si aucune chane n'existe
 **/
int Chaines::trouve_restauration_par_cle(UINT32 cleChaine) const
{
	int nRes = trouve_par_cle(cleChaine);

	if (nRes < 0 && !empty())
		nRes = 0;
	return nRes;
}

/**
 * Recherche d'une chane avec son jeu d'identificateurs complet
 * \param[in]	sIds	srie d'indentifiants de la chane
 * \retval		Index de la chane dans la table #Canaux
 * \retval		-1 si le numro n'a pas t trouv
 **/
int Chaines::trouve_par_Ids(const ChaineIDs & sIds) const
{
	int nbChaines = (int)size();

	for (int i=0; i<nbChaines; i++) {
		if ((*this)[i].EquIds(sIds))
			return i;
	}
	return -1;
}

/**
 * Dtermine le temps restant avant le prochain vnement EPG concernant n'importe quelle
 * chane.
 * \return		Temps restant, exprim en millisecondes, ou bien #DUREE_INFINIE si aucun temps
 *				ne peut tre dtermin (aucune mission dfinie dans l'EPG ...).
 **/
long Chaines::timeBeforeNextEvent()
{
	long		nTime	= DUREE_INFINIE;
	SYSTEMTIME	localtime;

	GetLocalTime(&localtime);
	for (ITERATE_CONST_VECTOR(*this, Chaine, it)) {
		long nTmp = it->timeBeforeNextEvent(&localtime);

		if (nTmp < nTime)
			nTime = nTmp;
	}
	return nTime;
}

/**
 * Trouver le SID correspodant  un index de chane.
 * \param[in]	ixChaine	Index dans la table #Canaux
 * \note Si on passe \p STREAM_PSEUDO_INDEX en paramtre, la fonction retourne \p STREAM_PSEUDO_SID.
 * Ceci est utilis en interne pour grer les enregistrements de multiplex entier.
 * \deprecated Remplacer par #indexToCle
 **/
WORD indexToSID(int ixChaine)
{
	return
		ixChaine==STREAM_PSEUDO_INDEX ? STREAM_PSEUDO_SID :
		ixChaine_ok(ixChaine) ? Canaux[ixChaine].SID : 0;
}

/**
 * Trouver la cl correspodant  un index de chane.
 * \param[in]	ixChaine	Index dans la table #Canaux
 * \note Si on passe \p STREAM_PSEUDO_INDEX en paramtre, la fonction retourne \p STREAM_PSEUDO_CLE.
 * Ceci est utilis en interne pour grer les enregistrements de multiplex entier.
 **/
UINT32 indexToCle(int ixChaine)
{
	return
		ixChaine==STREAM_PSEUDO_INDEX ? STREAM_PSEUDO_CLE :
		ixChaine_ok(ixChaine) ? Canaux[ixChaine].cle() : 0;
}

/**
 * Dconnexion globale des sorties vido et audio
 **/
bool suspend()
{
	static bool bProt = false;	// Protection contre la rentrance
	bool		bRes = false;

	if (!bProt) {
		bProt = true;

		if (!suspendu && ixChaine_ok(ixChaineCourante)) {
			myprintf(BOLD(TEXT("Dbranchement des sorties audio et vido")) EOL);

			HRESULT hr = Canaux[ixChaineCourante].suspend();

			if (FAILED(hr))
				affiche_erreurs(TEXT("Erreur lors du dbranchement des sorties"), hr);
			suspendu = true;
		}
		bRes = true;
		bProt = false;
	}

	return bRes;
}

/**
 * Reconnexion globale des sorties vido et audio
 * \param[in]	reset_son	\p true si le canal son doit en mme temps tre remis  sa
 *							valeur par dfaut, selon les prfrences globales de l'utilisateur
 **/
bool rebranche(bool reset_son)
{
	static bool bProt = false;	// Protection contre la rentrance
	bool		bRes = false;

	if (!bProt) {
		bProt = true;

		if (suspendu && ixChaine_ok(ixChaineCourante)) {
			myprintf(BOLD(TEXT("Branchement des sorties audio et vido")) EOL);
			suspendu = false;

			HRESULT hr = Canaux[ixChaineCourante].branche(reset_son);

			if (FAILED(hr))
				affiche_erreurs(TEXT("Erreur lors du rebranchement des sorties"), hr);
		}
		bRes = true;
		bProt = false;
	}

	return bRes;
}

/**
 * Changement de piste son
 * \param[in]	ixSon	Index de la nouvelle piste audio dans la table \p sons
 *						des pistes de la chane
 **/
void switch_audio(int ixSon)
{
	if (ixChaine_ok(ixChaineCourante)) {
		const Chaine &		cChan		= Canaux[ixChaineCourante];

		if (ixSon >= 0 && ixSon < cChan.nbr && ixSon != ixSonCourant) {
			ixSonCourant = ixSon;
			if (!suspendu) {
				cChan.branche(false);
				if (pOSD)
					pOSD->DisplaySoundInfo(cChan, ixSon);
			}
		}
	}
}

/**
 * Syntonisation d'une frquence dans le tuner.
 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
 * \param[in]	nNoCanal	Numro du canal  explorer (en principe de 21  69)
 * \param[in]	pChIDs		Autres informations optionnelles concernant la chane
 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
 **/
HRESULT change_frequence(LONG nFreq, LONG nBandwidth, int nNoCanal, const ChaineIDs * pChIDs, tstring * pstrErr)
{
	HRESULT	hr;
#if USE_CONSOLE
	tstring	strErr;

	if (!pstrErr)
		pstrErr = &strErr;
#endif

	hr = pGraph->change_frequence(nFreq, nBandwidth, pChIDs IF_CONSOLE_CB(&strErr));
	myprintf(CMyP ERRX(TEXT("Erreur(s) change_frequence : %s")), !pstrErr->empty(), pstrErr->c_str());

	// Notification  la fentre principale
	PostMessage(hMainWnd, WM_APP_NEW_FREQ, nNoCanal, nFreq);
	return hr;
}

/**
 * Changement de chane, slection par index dans la table.
 * \param[in]	ixChaine	Index dans la table #Canaux.
 *							(mettre -1 pour dsactiver la slection de chane)
 * \param[in]	bForceTune	Si \p true, le  tuner est resyntonis mme si la frquence
 *							ne change pas
 * \retval		S_OK si pas d'erreur
 **/
HRESULT change_chaine(int ixChaine, bool bForceTune)
{
	if (!pGraph)
		return E_FAIL;
	if ( !ixChaine_ok(ixChaine) && ixChaine != -1) {
		myprintf(ERRX(TEXT("change_chaine(%d) invalide")), ixChaine);
		return E_FAIL;
	}
	bool freq_differente	= true;
	bool chaine_differente	= ixChaine!=ixChaineCourante;

	if (!bForceTune) {
		if (ixChaine_ok(ixChaineCourante) && ixChaine_ok(ixChaine) &&
			Canaux[ixChaineCourante].sameMultiplexAndVideoType(Canaux[ixChaine]))
		{
			freq_differente = false;
		}

		if (freq_differente && enregistrements_actuels.recording())
			return S_FALSE;	// On ne change pas de frquence si enregistrement en cours
	}

	int ixSauveChaine = ixChaineCourante;

	ixChaineCourante = -1;	// doit tre gal  -1 pendant les transitions ( cause du filtre PMT)

	if (chaine_differente)
		debranche_medias();

	if (ixChaine >= 0) {
		const Chaine & nouveau_canal = Canaux[ixChaine];

		myprintf(
			TEXT("Zappe ix=%i, no=%i, SID=%i, TSID=%i, ONID=%i,") EOL
			TEXT("  freq=%i, nom=\"") COLR(HGREEN, TEXT("%") A2t) TEXT("\"") EOL,
			ixChaine, nouveau_canal.nNumeroChaine, nouveau_canal.SID, nouveau_canal.TSID,
			nouveau_canal.ONID, nouveau_canal.khz, nouveau_canal.nom);

		if (freq_differente)
			change_frequence(nouveau_canal.khz, DVBT_BANDWIDTH, nouveau_canal.canal_no, & nouveau_canal);

		if (ixChaine_ok(ixChaine)) {
			// Sauvegarde de la chane qu'on abandonne (pour zapping en va-et-vient)
			if (ixChaine_ok(ixSauveChaine) && ixSauveChaine != ixChaine)
				ixDerniereChaine = ixSauveChaine;
			ixChaineCourante = ixChaine;
		} else
			ixChaineCourante = -1;

		ixChaineAvance = ixChaineCourante;

		// Ne reconnecter l'interface que si elle n'tait pas suspendue :
		if (!suspendu && chaine_differente) {
			HRESULT hr = Canaux[ixChaineCourante].branche(true);

			if (FAILED(hr))
				affiche_erreurs(TEXT("Erreur lors du rebranchement des sorties"), hr);
		}
	}

	return S_OK;
}
