/*
 * $Id: filters.cpp 308 2011-11-24 11:09:11Z gingko $
 * Copyright (C) 2009 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation de fonctionnalits d'assistance  l'insertion des filtres
 *
 *	Ce fichier implmente des fonctionnalits facilitant la manipulation
 *	des filtres dans le graphe.
 **/

#include "stdafx.h"
#include "search.h"

// ====================================================================================
// Structures, dfinitions et mthodes concernant les filtres
// ====================================================================================

/**
 * \brief Cration du filtre
 **/
HRESULT	FilterData::Create()
{
	HRESULT hr = S_OK;	// "CreateInstance" est suppos n'assigner cette valeur qu'en cas d'erreur
						// (voir commentaire dans BaseClasses, fichier "combase.h").

	if (bLocal)
		pFilter = dynamic_cast<IBaseFilter *>(procCrea(NULL, &hr));
	else if (pclsid)
		hr = pFilter.CoCreateInstance(*pclsid, NULL, CLSCTX_INPROC_SERVER);

	return hr;
}

/// Cration et ajout du filtre.
HRESULT	FilterData::get_and_add(IPTvM_GraphBase & iGraph)
{
	HRESULT hr = S_OK;

	if (!pFilter)
		hr = Create();

	if (FAILED(hr))
		return erreur(TEXT("Le filtre \"%s\" n'a pas pu tre cr"), hr, pszName);

	myprintf(TEXT("Ajout filtre \"%s\" (0x%x)") EOL, pszName, &*pFilter);

	hr = iGraph.AddFilterToGraph(pFilter, pszName);	// Code retour 0x0004022d (VFW_S_DUPLICATE_NAME)
													// possible, mais ce n'est pas une erreur.
	if (FAILED(hr))
		return erreur(TEXT("Le filtre \"%s\" n'a pas pu tre insr dans le graphe"), hr, pszName);
	return hr;
}

/**
 * \brief Appel du dialogue de proprits du filtre
 **/
VOID FilterData::create_property_dialog()
{
	if (!pFilter) {
		affiche_erreurs(TEXT("Il faut d'abord configurer un filtre et redmarrer l'application"));
		return;
	}

	CComQIPtr<ISpecifyPropertyPages> pPages(pFilter);

	if (!pPages) {
		// (avertissement maintenant en principe inutile car dans ce cas, le menu est dsactiv)
		affiche_erreurs(TEXT("Ce filtre n'a pas de page de proprits"), E_NOINTERFACE);
		return;
	}

	CComPtr<IUnknown> pUnk(pPages);

	if (pUnk) {
		CAUUID		caGUID;
		HRESULT		hr = pPages->GetPages(&caGUID);

#ifdef _UNICODE
		LPOLESTR	pszTmp	= LPOLESTR(pszName);
#else
		COPY_STR_ON_STACK(OLECHAR, pszTmp, pszName, nLen);
#endif
		if (SUCCEEDED(hr)) {
			LCID lcid = GetUserDefaultLCID();

			ShowCursor(TRUE);

			hr = OleCreatePropertyFrame(
					hMainWnd,
					10,
					10,
					pszTmp,
					1,
					&pUnk.p,
					caGUID.cElems,
					caGUID.pElems,
					lcid,
					0L,
					NULL );

			ShowCursor(FALSE);
			CoTaskMemFree(caGUID.pElems);

			if (FAILED(hr)) {
				affiche_erreurs(TEXT("Erreur lors de la cration de la page de proprits"), hr);
			}
		}
	}
}

/**
 * \brief Vrification de l'existence d'un dialogue de proprits sur ce filtre
 *
 * \return	\p true si un dialogue de proprits existe
 * \note	La mthode cherchera  crer une instance du filtre s'il n'y en a pas encore
 **/
bool FilterData::check_property_dialog()
{
	if (isNull())
		return false;
	if (!pFilter) {
		HRESULT hr = Create();

		check_hr(hr, "check_property_dialog / Create");
	}
	if (!pFilter)
		return false;
	CComQIPtr<ISpecifyPropertyPages> pPages(pFilter);

	return !!pPages;
}

/**
 * \brief Obtenir un pointeur sur le graphe dans lequel est insr un filtre
 * \return Pointeur sur le graphe, ou bien NULL si pas de filtre ou filtre non insr
 **/
CComPtr<IFilterGraph> FilterData::graph_from_filter()
{
	CComPtr<IFilterGraph>	pGrph;

	if (pFilter) {
		FILTER_INFO	sInfo;
		HRESULT		hr = pFilter->QueryFilterInfo(&sInfo);

		if (hr == S_OK && sInfo.pGraph)
			// Le pointeur a dj une rfrence compte, donc on ne doit pas en ajouter
			pGrph.p = sInfo.pGraph;
	}
	return pGrph;
}

HRESULT DemuxOutPin::CreatePin(IBaseFilter * pDemuxFilter, const AM_MEDIA_TYPE & sAm, bool bFindExisting)
{
	HRESULT hr;

	if (bFindExisting) {
		hr = cherche_pin(pDemuxFilter, PINDIR_OUTPUT, pPin, &sAm);
		if (hr != E_FAIL) {
			myprintf(CMyP TEXT("Broche existante trouve pour \"") BOLD(TEXT("%") W2t) TEXT("\"") EOL,
				SUCCEEDED(hr), pszName);
			return hr;
		}
	}

	CComQIPtr<IMpeg2Demultiplexer>	pMpeg2Demux(pDemuxFilter);

	if (!pMpeg2Demux)
		return erreur(TEXT("L'interface du dmultiplexeur n'a pas t trouve"), E_NOINTERFACE);

	// Note : utilisation de "const_cast" ici car on ne s'attend pas normalement  ce que
	// "CreateOutputPin" modifie les objets points.
	hr = pMpeg2Demux->CreateOutputPin(const_cast<AM_MEDIA_TYPE *>(&sAm),
								      const_cast<LPWSTR>(pszName), &pPin);
	myprintf(CMyP TEXT("Broche \"") BOLD(TEXT("%") W2t) TEXT("\" cre") EOL, SUCCEEDED(hr), pszName);
	if (FAILED(hr))
		return erreur(TEXT("La broche \"%") A2t TEXT("\" n'a pas pu tre cre"), hr, pszName);
	if (FAILED(hr = pPin.QueryInterface(&pMap)))
		return erreur(TEXT("L'interface de la broche \"%") A2t TEXT("\" n'a pas t trouve"), hr, pszName);
	return hr;
}

/**
 * Branchement d'un flux lmentaire (identifi par PID) sur la connexion \p pMap
 * (en sortie du dmultiplexeur), en dbranchant tous les autres ventuels
 * \param[in]	pid		N de PID  y connecter, ou 0 si aucun
 * \param[in]	bKeep	Ne dbranche PAS les autres PIDs si \p true
 * \return	Code d'erreur DirectShow
 * \note	Utiliser 0 comme paramtre \p pid pour tout dbrancher
 **/
HRESULT DemuxOutPin::Map(ULONG pid, bool bKeep) const
{
	HRESULT hr = E_FAIL;

	if (pMap) {
		CComPtr<IEnumPIDMap>	pEnumMap;

		if (SUCCEEDED(hr = pMap->EnumPIDMap(&pEnumMap))) {
			PID_MAP				maps[8];
			ULONG				nbPids = 0;
			bool				bNeedMap = pid != 0;

			if (pEnumMap->Next(_countof(maps), maps, &nbPids)==S_OK) {
				for (ULONG i=0; i<nbPids; i++) {
					ULONG		oldpid = maps[i].ulPID;

					if (pid != 0 && oldpid == pid) {	// Le PID  connecter est dj connect
						bNeedMap = false;
						continue;
					}

					if (!bKeep) {
						myprintf(TEXT("Dbranche pid=") BOLD(TEXT("%u")) TEXT(" de la broche ") BOLD(TEXT("%") W2t) EOL, oldpid, pszName);
						if (FAILED(hr = pMap->UnmapPID(1, &oldpid))) {
							myprintf(ERRX(TEXT("chec UnmapPID(%u) (%") W2t TEXT("), code 0x%08x")), oldpid, pszName, hr);
							break;
						}
					}
				}
			}

			if (bNeedMap) {
				myprintf(TEXT("Branche pid=") BOLD(TEXT("%u")) TEXT(" sur broche ") BOLD(TEXT("%") W2t) EOL, pid, pszName);
				hr = pMap->MapPID(1, &pid, eMsc);
				myprintf(CMyP ERRX(TEXT("chec branchement pid=%u sur broche %") W2t TEXT(", code 0x%08x")), FAILED(hr), pid, pszName, hr);
			}
		}
	}
	return hr;
}

/// Tester si l'application d'un codec et d'un rendu ncessitent un changement du graphe
bool MediaOutPin::NeedGraphChange(const DecoderData * pNewCodec, RdrTypes eNewVmrMode) const
{
	if (!pNewCodec)
		eNewVmrMode = rdt_null;	// pas de rendu sans codec

	bool bSameVmr	= pRenderer && eNewVmrMode == pRenderer->GetRdrType();
	bool bSameCodec	= pNewCodec == pCurCodec;

	return ! (bSameVmr && bSameCodec);
}

/**
 * Slection, avec cration ventuelle, d'un rendu vido
 * \param[in]	iGraph			Graphe dans lequel slectionner le rendu vido
 * \param[in]	bNoRendering	Si \p true, remplacer par un non rendu vido (croix bleue)
 * \return		Code d'erreur DirectShow
 **/
HRESULT	VideoOutPin::SelectRenderer(IPTvM_GraphBase & iGraph, bool bNoRendering)
{
	RdrTypes eRdrMode = bNoRendering ? rdt_null : vmr_mode;

	// Dtruire le rendu vido prcdent s'il en existait un
	if (pRenderer) {
		if (pRenderer->GetRdrType() == eRdrMode)
			return S_OK;	// Mme rendu que prcdemment : pas de changement

		pRenderer.reset();
	}

	HRESULT hr = E_FAIL;

	pVMR.reset();
	pRenderer.reset(CVideoRenderer::NewRenderer(eRdrMode));
	if (pRenderer)
		hr = pRenderer->Init(iGraph);
	if (FAILED(hr)) {
		// Une erreur  l'initialisation du VMR n'est pas interruptive : le reste du
		// graphe doit pouvoir continuer  fonctionner (nanmoins, inutile de tenter
		// d'insrer le codec vido aprs cela).
		erreur(TEXT("Erreur lors de la slection du rendu vido"), hr);
		affiche_erreurs(
			TEXT("Le module de rendu vido (VMR) n'a pas pu s'initialiser. Le programme va s'excuter") EOL
			TEXT("avec l'affichage vido dsactiv. Merci de vous assurer de la disponibilit de votre") EOL
			TEXT("carte graphique, puis de vous rendre dans la configuration de Pouchin TV Mod") EOL
			TEXT("(onglet \"Options\") pour y choisir et y (r)activer une mthode de rendu vido."), S_OK);
		pRenderer.reset(CVideoRenderer::NewRenderer(rdt_null));
		if (pRenderer)
			hr = pRenderer->Init(iGraph);
	}

	// Donner un accs en variable globale  cet objet (pas trs lgant, mais plus
	// simple, et pour le moment on va continuer  s'en contenter)
	pVMR = dynamic_pointer_cast<CVideoRenderer>(pRenderer);
	return hr;
}

VideoOutPin::~VideoOutPin()
{
	// Supprimer le pointeur global sur le rendu vido
	pVMR.reset();
}

/**
 * Slection, avec cration ventuelle, d'un rendu audio
 * \param[in]	iGraph			Graphe dans lequel slectionner le rendu audio
 * \param[in]	bNoRendering	Si \p true, remplacer par un rendu audio NULL
 * \return		Code d'erreur DirectShow
 **/
HRESULT	AudioOutPin::SelectRenderer(IPTvM_GraphBase & iGraph, bool bNoRendering)
{
	RdrTypes eRdrMode = bNoRendering ? rdt_null : rdt_DirectSound;

	// Dtruire le rendu audio prcdent s'il en existait un
	if (pRenderer) {
		if (pRenderer->GetRdrType() == eRdrMode)
			return S_OK;	// Mme rendu que prcdemment : pas de changement

		pRenderer.reset();
	}

	HRESULT hr = E_FAIL;

	pRenderer.reset(CAudioRenderer::NewRenderer(eRdrMode));

	if (pRenderer)
		hr = pRenderer->Init(iGraph);
	if (FAILED(hr))
		return erreur(TEXT("Erreur lors de la slection du rendu audio"), hr);
	return hr;
}

HRESULT DecoderData::render(IPTvM_GraphBase & iGraph, FilterData * pFilterOut)
{
	HRESULT hr = get_and_add(iGraph);

	if (SUCCEEDED(hr)) {
		hr = iGraph.connect_pin_to_filter(cDemuxOut.pPin, pFilter);
		if (hr != S_OK) {
			erreur(
				TEXT("Impossible de connecter le dmultiplexeur au filtre %") A2t TEXT(" \"%s\"."),
				hr, pszType, pszName);
		} else if (pFilterOut) {
			hr = iGraph.connect_filters(pFilter, pFilterOut->pFilter);
			if (hr != S_OK) {
				erreur(
					TEXT("Impossible de connecter le filtre %") A2t TEXT(" \"%s\" au rendu \"%s\"."),
					hr, pszType, pszName, pFilterOut->pszName);
			}
		}
	}
	if (FAILED(hr)) {
		affiche_erreurs(NULL, hr);
		// Insertion ou connexion du codec impossible : retrait du codec
		HRESULT hr2 = iGraph.remove_filter(pFilter);

		check_hr(hr2, "invalid codec / remove_filter");
	}
  #if HOOK_INTERFACES
	init_IMPEG2PIDMap_hook(cDemuxOut.pMap);
  #endif
	return hr;
}

/**
 * \brief Cration du filtre
 **/
HRESULT MediaDecoderData::Create()
{
	HRESULT hr = S_OK;

	if (!pFilter)
		hr = CSearchByType_Get(pAmt->majortype, pAmt->subtype, exact_match, pszName, pFilter).Do();
	return hr;
}

HRESULT MediaDecoderData::render(IPTvM_GraphBase & iGraph)
{
	if (isNull())
		return S_OK;

	HRESULT	hr = cDemuxOut.SelectRenderer(iGraph, false);

	if (FAILED(hr))
		return hr;
	return __super::render(iGraph, cDemuxOut.GetRenderer());
}

// ====================================================================================
// Assistance  la recherche de filtres
// ====================================================================================

// Structure d'assistance  la rcupration d'une chane de caratres depuis une variable de type
// VARIANT (permet en particulier de nettoyer la chane au passage, pour corriger un problme
// constat avec certains tuners):
struct VarStr_helper
{
	TCHAR str[128];

	/// Constructeur
	VarStr_helper(const VARIANT & var);

	/// Rcupration du rsultat
	LPCTSTR operator () () const {
				return str;}
};

/// Constructeur
VarStr_helper::VarStr_helper(const VARIANT & var)
{
	if (var.vt==VT_BSTR) {
		strcpy_T(str, var.bstrVal);
		str[_countof(str)-1] = 0;	// prcaution

		size_t len = strlen_T(str);

		while (len>0 && str[--len]<=' ')
			str[len] = 0;
	} else
		str[0] = 0;
}

/**
 * Constructeur  partir d'une numration.
 * \param[in]	pMonikr	Pointeur vers l'interface "Moniker" numre
 * \param[in]	phr		Pointeur sur variable o mettre le code retour si non NULL
 **/
MonikerBindings::MonikerBindings(IMoniker * pMonikr, HRESULT * phr) :
	pMoniker(pMonikr)
{
	HRESULT hr = CreateBindCtx(0, &pBindCtx);

	if (!check_hr(hr, "CreateBindCtx"))
		hr = BindProperties();

	if (phr)
		*phr = hr;
}

class CErrorLog : public IErrorLog
{
public:
	// Implmentation triviale des mthodes (abstraites) de l'interface \p IUnknown
	virtual STDMETHODIMP QueryInterface(REFIID riid, __deref_out void **ppv) {
		return NULL; }

	virtual STDMETHODIMP_(ULONG) AddRef() {
		return 0; }

	virtual STDMETHODIMP_(ULONG) Release() {
		return 0; }

	virtual HRESULT STDMETHODCALLTYPE AddError(LPCOLESTR pszPropName, EXCEPINFO *pExcepInfo) {
		return S_OK; }
};

/// Rcupration \p IPropertyBag et nom du filtre
HRESULT	MonikerBindings::BindProperties()
{
	CErrorLog	cErrLog;

	pPropBag.Release();	// Au cas o on appellerait une seconde fois cette mthode

	HRESULT hr = pMoniker->BindToStorage(pBindCtx, 0, IID_IPropertyBag, reinterpret_cast<void**>(&pPropBag));

	if (!check_hr(hr, "BindToStorage")) {
		CComVariant varName;

		hr = pPropBag->Read(L"FriendlyName", &varName, &cErrLog);
		if (!check_hr(hr, "lecture nom du filtre"))
			strcpy_T(strName, VarStr_helper(varName)());
	}

	return hr;
}

/**
 * Charger l'interface du filtre dans \p pFilter (non ralis  la construction car pas
 * toujours requis et parfois gnant).
 * \return	Code d'erreur DirectShow
 **/
HRESULT MonikerBindings::BindToFilter()
{
	HRESULT hr = pMoniker->BindToObject(pBindCtx, NULL, IID_IBaseFilter, reinterpret_cast<void**>(&pFilter));

	check_hr(hr, "BindToObject");
	return hr;
}

MonikerBindings::MonikerBindings(LPCTSTR pszDevicePath, HRESULT * phr)
{
	HRESULT hr = CreateBindCtx(0, &pBindCtx);

	if (SUCCEEDED(hr)) {
		// Enlever les guillemets de part et d'autre, le cas chant
		// (ces guillemets avaient t ajouts pour prserver les ventuelles espaces en fin de chane,
		// lesquels n'auraient autrement pas t prservs dans le fichier de configuration)
		OLECHAR	szWrk[256];
		bool	bHasQuotes = pszDevicePath[0] == TCHAR('[');

		strcpy_T(szWrk, pszDevicePath + (bHasQuotes ? 1 : 0));
		if (bHasQuotes) {
			DWORD	nLen = strlen_T(szWrk);

			while (nLen>0 && szWrk[nLen-1]==TCHAR(' '))
				szWrk[--nLen] = 0;	// (un risque existe qu'il y ait quand mme aussi des espaces ajouts ...)
			while (nLen>0 && szWrk[nLen-1]==TCHAR(']'))
				szWrk[--nLen] = 0;
		}

		DWORD	dwEaten;

		hr = MkParseDisplayName(pBindCtx, szWrk, &dwEaten, &pMoniker);
		if (!check_hr(hr, "lecture device path") && pMoniker)
			hr = BindProperties();
	}

	if (phr)
		*phr = hr;
}

/**
 * Rcuprer le "device path" dans une variable \p tstring. Le "device path" contient  lui seul
 * toutes les informations ncessaires pour identifier un filtre existant de faon unique.
 * \param[out]	strDevicePath	"Device path" rcupr
 * \return		Code d'erreur DirectShow
 **/
HRESULT MonikerBindings::GetDevicePath(tstring & strDevicePath)
{
	LPOLESTR		pszDisplayName;
	HRESULT			hr = pMoniker->GetDisplayName(pBindCtx, NULL, &pszDisplayName);

	if (SUCCEEDED(hr)) {
		strcpy_T(strDevicePath, pszDisplayName);

		// Libration de la mmoire alloue pour le nom
		IMalloc *	memAlloc = NULL;

		CoGetMalloc(1, &memAlloc);
		memAlloc->Free(pszDisplayName);
		/// \todo IMalloc->Release() ???
	} else
		myprintf(ERRX(TEXT("Impossible de rcuprer le 'device path' du filtre \"%s\", hr=0x%08x")), strName.c_str(), hr);

	return hr;
}

// ====================================================================================
// Recherche de broche dans un filtre
// ====================================================================================

/**
 * Vrification de la compatibilit d'une broche avec un type de mdia donn.
 * \param[in]	pPin	Interface de la broche  tester
 * \param[in]	am		Descripteur de type de mdia
 * \retval \p true en cas de succs
 **/
bool check_pin_for_media(CComPtr<IPin> & pPin, const AM_MEDIA_TYPE & am)
{
	CComPtr<IEnumMediaTypes>	pEnumMedia;
	AM_MEDIA_TYPE *				pmt;
	bool						found = false;

	for (
		pPin->EnumMediaTypes(&pEnumMedia);
		!found && (pEnumMedia->Next(1, &pmt, NULL))==S_OK;
		DeleteMediaType(pmt)
	) {
		if (am.majortype == pmt->majortype && am.subtype == pmt->subtype)
			found = true;
	}
	//if (pPin->QueryAccept(&am)==S_OK) {
	//	return true;
	//}
	return found;
}

/**
 * Recherche de la premire broche d'un filtre, pour une direction donne.
 * \param[in]	pFilter	Interface du filtre dans lequel la recherche s'effectue
 * \param[in]	dir		Direction de la broche recherche (entre ou sortie)
 * \param[out]	result	Interface de la broche trouve
 * \param[in]	pmt		Si != \p NULL, alors la broche doit tre non connecte, et si la valeur
 *						est diffrente de NOTCONNECTED, elle doit en plus accepter le mdia spcifi
 * \retval \p S_OK	en cas de succs
 **/
HRESULT cherche_pin(IBaseFilter * pFilter, PIN_DIRECTION dir,
					CComPtr<IPin> & result, const AM_MEDIA_TYPE * pmt)
{
	// myprintf(TEXT("** cherche_pin") EOL);
	CComPtr<IEnumPins>	pEnumPins;
	HRESULT				hr = pFilter->EnumPins(&pEnumPins);

	if (FAILED(hr)) {
		myprintf(FMT_ERRX("cherche_pin/EnumPins(%i)"), dir, hr);
		return hr;
	}

	for (ENUM_ITF(pEnumPins, IPin, pPin)) {
		PIN_DIRECTION direction;

		if (FAILED(hr = pPin->QueryDirection(&direction))) {
			myprintf(FMT_ERRX("cherche_pin/QueryDirection(%i)"), dir, hr);
			break;
		}

		if (direction == dir) {
			if (pmt) {
				// write_pin_info(pPin);
				CComPtr<IPin> connected;

				if ((hr = pPin->ConnectedTo(&connected)) == VFW_E_NOT_CONNECTED) {
					hr = S_OK;
					if (pmt==NOTCONNECTED || check_pin_for_media(pPin, *pmt)) {
						result = pPin;
						break;
					}
				} else if (FAILED(hr)) {
					myprintf(FMT_ERRX("cherche_pin/ConnectedTo(%i)"), dir, hr);
					break;
				}
			} else {
				result = pPin;
				break;
			}
		}
	}

	return !result ? E_FAIL : hr;
}
