/*
 * $Id: osd_objects.cpp 392 2020-10-06 21:18:34Z gingko $
 * Copyright (C) 2011 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentations pour les objets OSD.
 *
 *	Implmentations des fonctionnalits ncessaires  l'affichage sur cran (OSD).
 **/

#include "stdafx.h"
#include "rendering.h"

// ====================================================================================

/// Obtenir les dimensions d'une image Bitmap  partir de sa "handle"
SIZE GetBmpSize(HBITMAP hBmp)
{
	BITMAP	sBitmap;
	SIZE	sSiz = {0};

	if (hBmp && GetObject(hBmp, sizeof(sBitmap), &sBitmap) >= sizeof(sBitmap)) {
		sSiz.cx	= sBitmap.bmWidth;
		sSiz.cy = sBitmap.bmHeight;
	}
	return sSiz;
}

// ====================================================================================
// Implmentations des mthodes de la classe COsdObjectText
// ====================================================================================

/// criture d'un nouveau texte dans ce rectangle OSD
/// \param[in]	pszFmt	Chane format style \p printf
/// \param[in]	argptr	Liste d'arguments, selon format
void COsdObjectText::SetText(LPCTSTR pszFmt, va_list argptr)
{
	Clear();		// Effacement de l'ventuel texte prcdent

	// Copie du texte format dans le tampon. Le texte est tronqu s'il est trop long.
	nTextSize = _vsntprintf_s(szText, _countof(szText)-1, pszFmt, argptr);
	szText[nTextSize] = 0;
}

COsdFontAllocatorPtr COsdObjectText::GetFontAllocator() const
{
	return cOSD.pOsdFontAlloc;
}

/// Effacement dfinitif des donnes de cet objet OSD.
void COsdObjectText::Clear()
{
	if (nTextSize) {
		nTextSize = 0;
		szText[0] = 0;
	}
}

// ====================================================================================

COsdObjectBmp::~COsdObjectBmp()
{
	if (m_hDC)
		DeleteDC(m_hDC);
	if (m_hBmp && m_hBmp != INVALID_HANDLE_VALUE)
		DeleteBitmap(m_hBmp);
}

// ====================================================================================
// Implmentations des mthodes de la classe COsdAllObjects
// ====================================================================================

/// Constructeur
COsdAllObjects::COsdAllObjects() :
	last_event(0),
	pOsdFontAlloc(pVMR ? pVMR->CreateFontAllocator() : COsdFontAllocator::CreateInstance(lt_GDI)),
	bEnabled(false)
{
	ZeroMemory(&sOsdRect, sizeof(sOsdRect));
}

UINT COsdAllObjects::GetFontSize() const
{
	return pOsdFontAlloc ? pOsdFontAlloc->Size() : 0;
}

/// Calcul du rectangle utilis pour la vido ( l'usage de la construction de l'OSD)
/// \param[in] sRect	Rectangle client de la fentre, calcul par \p CalcCliRect
/// \return				Rectangle calcul, utilisable pour l'OSD
RECT COsdAllObjects::CalcOsdRect(RECT sRect)
{
	// myprintf(COLR(HGREEN, TEXT("COsdAllObjects::CalcOsdRect, pVMR=0x%08x")) EOL, pVMR);
	// Si le VMR existe, on lui dlgue cette tche
	if (pVMR)
		return pVMR->CalcOsdRect(sRect);

	// Sinon, on affiche en mode texte dans la fentre (noire) selon des valeurs standard
	static SIZE sAr = {DEFAULT_AR_X, DEFAULT_AR_Y};

	PlaceVideoInRect(sRect, sAr, &sRect);
	return sRect;
}

/// Calculer la hauteur de la police  utiliser
/// \param[in] sRect Rectangle OSD de la fentre, calcul par \p CalcOsdRect
LONG COsdAllObjects::CalcFontHeight(const RECT & sRect)
{
	LONG	nHgt = RHgt(sRect);	// Hauteur de l'affichage
	LONG	nWdt = RWdt(sRect);	// Largeur de l'affichage
	float	fDim = float((nHgt*nHgt)+(nWdt*nWdt));

	// On essaie d'viter d'avoir une hauteur de police infrieure  32 (possible en cas
	// de fort zoom)
	LONG	nFHgt = max(32, LONG(sqrt(fDim))/24);

	// ... mais nanmoins, la hauteur ne saurait tre infrieure au tiers de l'affichage
	// (cas possible si la vido est force en largeur et que la fentre est trs aplatie)
	return min(nHgt/3, nFHgt);
}

/// Effacement de tous les objets OSD affichs
void COsdAllObjects::Clear()
{
	BeginChange();
	setOsdObjects.clear();
	EndChange();
	last_event = 0;
}

/// Mthode  appeler en rponse  l'expiration des dlais lis  l'OSD
void COsdAllObjects::ProcessTimeEvent()
{
	KillTimer(hMainWnd, TIMER_OSD);

	if (last_event != 0) {
		DWORD dwRemain = GetTickCount() - last_event;

		if (dwRemain >= DUREE_OSD) {
			// Expiration de la dure d'affichage OSD :
			// On efface les donnes sur tous les objets affichs
			Clear();
		} else {
			// Un nouvel vnement a t programm, ou bien, pour une raison quelconque,
			// on a reu un vnement alors que le dlai n'est pas expir :
			SetTimer(hMainWnd, TIMER_OSD, dwRemain, NULL);
		}
	}
}

void COsdAllObjects::SetOsdObject(COsdObject * pNewObj)
{
	COsdObjectPtr	pOsdObj	= COsdObjectPtr(pNewObj);
	COsdObjectIter	ite		= setOsdObjects.end();
	COsdObjectIter	itf		= setOsdObjects.find(pOsdObj);

	if (itf == ite) {
		// L'objet OSD correspondant n'existe pas
		if (pOsdObj->HasContents())
			setOsdObjects.insert(pOsdObj);
	} else {
		// L'objet OSD correspondant existe dj
		if (pOsdObj->HasContents())
			swap((COsdObjectPtr &)pOsdObj, (COsdObjectPtr &)*itf);
		else
			setOsdObjects.erase(pOsdObj);
	}
}

void COsdAllObjects::SetOsdText(COsdObjectText * pNewObj, LPCTSTR pszFmt, va_list argptr)
{
	if (pszFmt && pNewObj)
		pNewObj->SetText(pszFmt, argptr);
	SetOsdObject(pNewObj);
}

/// Redessiner tous les objets dfinis
/// \return	\p true si au moins un texte a t raffich
bool COsdAllObjects::Redraw(HDC hDC)
{
	if (!bEnabled)
		return false;

	if (pVMR) {
		COsdLayerPtr pOsdLayer = pVMR->BeginDraw(hDC);

		if (pOsdLayer) {
			if (bEnabled) {
				for (COsdObjectIter it = setOsdObjects.begin(); it != setOsdObjects.end(); ++it) {
					if ((*it)->HasContents())
						pOsdLayer->DrawOsdObject(*it);
				}

				if (pOsdLayer->notEmpty()) {
					last_event = GetTickCount();
					ProcessTimeEvent();
				}
			}

			pVMR->EndDraw();
			return pOsdLayer->notEmpty();
		}
	}
	return false;
}

/// Dsactivation de l'affichage OSD : tout message reu pendant la dsactivation
/// sera mis en attente et affich  la ractivation.
void COsdAllObjects::Disable()
{
	myprint(COLR(HBLUE, TEXT("COsdAllObjects::Disable()")) EOL);
	bEnabled = false;
}

/// Rsactivation de l'affichage OSD, avec affichage effectif des messages qui
/// sont en attente.
void COsdAllObjects::Enable()
{
	UpdateRectsAndFont();

	myprint(COLR(HBLUE, TEXT("COsdAllObjects::Enable()")) EOL);
	if (! bEnabled) {
		bEnabled = true;
		Redraw(NULL);
	}
}

void COsdAllObjects::BeginChange()
{
	// Vide ... pour le moment
}

void COsdAllObjects::EndChange()
{
	if (bEnabled)
		Redraw(NULL);
}

/// Vrifie si le rectangle de l'OSD a chang. Retourne \p true si oui. Optionnellement,
/// place le nouveau rectangle dans la variable pointe
bool COsdAllObjects::CheckOsdRectChange(LPRECT pNewRect)
{
	// (re)Calculer le rectangle de l'OSD :
	RECT sNewOsdRect = CalcOsdRect(CalcCliRect());

	if (sNewOsdRect == sOsdRect)
		return false;
	if (pNewRect)
		*pNewRect = sNewOsdRect;
	return true;
}

/// Actualisation des rectangles et de la police d'affichage
///  utiliser  chaque redimensionnement de la vido
void COsdAllObjects::UpdateRectsAndFont()
{
	// (re)Calculer le rectangle de l'OSD :
	RECT sNewOsdRect;

	if (CheckOsdRectChange(&sNewOsdRect)) {
		//myprintf(COLR(LYELLOW, TEXT("Mise  jour des coordonnes OSD : Old=%i%i, New=%i%i")) EOL,
		//	RWdt(sOsdRect), RHgt(sOsdRect), RWdt(sNewOsdRect), RHgt(sNewOsdRect));

		// Masquer tous les objets dfinis :
		// Hide();	// Supprim (dplac dans 'update_coords')

		sOsdRect = sNewOsdRect;

		// (re)Calculer la hauteur de la police :
		LONG nHgt = CalcFontHeight(sOsdRect);

		// Si la hauteur de la police a chang, dfinir une nouvelle taille de police :
		pOsdFontAlloc->SetSize(nHgt);

		// Redessiner  nouveau tous les objets dfinis, selon leur nouvelle position :
		Redraw(NULL);
	}
}

/// Affichage du nom de la chane et du nom de l'mission en cours
/// \param[in] canal Descripteur de la chane courante
void COsdAllObjects::DisplayChannelInfo(const Chaine & canal)
{

	BeginChange();
	SetOsdText<COsdObjectTopRight>(TEXT("%") A2t, canal.nom);
	SetOsdText<COsdObjectTopRightL2>(TEXT("%s"), canal.getCurrentProgram().getName(TEXT("...")).c_str());
	EndChange();
}

/// Affichage d'un numro de chane (affichage en cours de saisie)
/// \param[in] nChanNumber Numro  afficher
void COsdAllObjects::DisplayChannelNumber(WORD nChanNumber)
{
	BeginChange();
	SetOsdText<COsdObjectTopRight>(TEXT("%i"), nChanNumber);
	EndChange();
}

/// Affichage du volume sonore
/// \param[in] nCurrentVolume niveau de volume courant
void COsdAllObjects::DisplayVolume(long nCurrentVolume)
{
	BeginChange();
	SetOsdText<COsdObjectBottomLeft>(TEXT("Volume %i"), nCurrentVolume);
	EndChange();
}

/// Affichage du zoom
/// \param[in] nZoomRatio zoom courant
void COsdAllObjects::DisplayZoom(int nZoomRatio)
{
	BeginChange();
	SetOsdText<COsdObjectBottomLeft>(TEXT("Zoom %i%%"), nZoomRatio);
	EndChange();
}

/// Affichage du nom d'une piste son
/// \param[in] canal	Descripteur de la chane courante
/// \param[in] nTrack	Numro de la piste son
void COsdAllObjects::DisplaySoundInfo(const Chaine & canal, int nTrack)
{
	TCHAR	tab[64];

	canal.toTrackMenuString(nTrack, tab, _countof(tab), nTrack+1);
	BeginChange();
	SetOsdText<COsdObjectBottomRight>(TEXT("%s"), tab);
	EndChange();
}

/// Affichage de l'tat "mute"
/// \param[in] bMuted true si on est en mode silencieux
void COsdAllObjects::DisplayMute(bool bMuted)
{
	BeginChange();
	SetOsdObject(new COsdObjectSoundBmp(*this, bMuted ? IDB_MUTE_ON : IDB_MUTE_OFF));
	EndChange();
}

/// Affichage d'un message quelconque (dans le coin infrieur droit)
void COsdAllObjects::DisplayBotRightMsg(LPCTSTR pszMsg)
{
	BeginChange();
	SetOsdText<COsdObjectBottomRight>(pszMsg);
	EndChange();
}

// ====================================================================================

/// Constructeur
/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
COsdObjectSoundBmp::COsdObjectSoundBmp(COsdAllObjects & cOsd, UINT nBmpId) :
	COsdObjectBmp(cOsd, osd_Sound)
{
	m_hBmp = static_cast<HBITMAP>(
		LoadImage(hAppInstance, MAKEINTRESOURCE(nBmpId), IMAGE_BITMAP, 0, 0, LR_MONOCHROME));

	if (!m_hBmp)
		return;
	m_hDC = CreateCompatibleDC(AutoDC(hMainWnd));
	myprintf(CMyP TEXT("Erreur CreateCompatibleDC, code=0x%08x") EOL, m_hDC == NULL, GetLastError());
}

// ====================================================================================
