/*
 * $Id: osd_objects.h 392 2020-10-06 21:18:34Z gingko $
 * Copyright (C) 2011 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions pour les objets OSD.
 *
 *	Dfinitions des classes et structures ncessaires  l'affichage sur cran (OSD).
 **/

#pragma once

#include <set>

// ====================================================================================
// Dfinitions
// ====================================================================================

/// Dlai de persistance des affichages OSD
#define DUREE_OSD			3000			// Valeur en millisecondes

enum OsdObjType
{
	oot_Text,
	oot_Bmp
};

/// numration dfinissant l'ordre d'affichage des objets OSD par leur valeur numrique
/// (valeurs les plus faibles en premier), et les identifiant galement de faon unique.
enum OsdLabelsAndZOrder {
	osd_Top,		//!< Textes OSD du haut
	osd_Top2,		//!< Textes OSD 2me rang du haut
	osd_BotRight,	//!< Textes OSD du bas  droite
	osd_BotLeft,	//!< Textes OSD du bas  gauche
	osd_Sound		//!< Bitmap "Muet"
};

// ====================================================================================

/// Obtenir les dimensions d'une image Bitmap  partir de sa "handle"
SIZE GetBmpSize(HBITMAP hBmp);

// ====================================================================================
// Dclaration de la classe COsdObject
// ====================================================================================


/// Classe de base pour les objets OSD.
class COsdObject
{
	friend class COsdAllObjects;
	friend class COsdLayer;
	friend struct OsdLess;	// functor

protected:

	COsdAllObjects	&	cOSD;			//!< Rfrence vers l'objet qui contient celui-ci
	OsdLabelsAndZOrder	m_eZOrder;		//!< Ordre d'affichage, et cl de recherche pour effacement

	/// Fonction virtuelle rajustant les rfrences de l'objet OSD, pralablement
	/// initialis avec des valeurs communes, pour les rduire  celles de l'objet
	/// spcifique courant.
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void		AdjustOsdRect(RECT & sRct) const = 0;

public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets
	/// \param[in]	eZOrder	Ordre d'affichage, et cl de recherche pour effacement
	COsdObject(COsdAllObjects & cOsd, OsdLabelsAndZOrder eZOrder) :
		cOSD(cOsd),
		m_eZOrder(eZOrder)
	{
	}

	/// Retourne le type de l'objet OSD
	virtual OsdObjType	Type() const = 0;

	/// Tester si un contenu existe
	virtual bool		HasContents() const = 0;

	/// Effacement dfinitif des donnes de cet objet OSD.
	virtual void		Clear() = 0;

	virtual ~COsdObject() {
				}
};

// ====================================================================================
// Dclaration de la classe COsdObjectText
// ====================================================================================

typedef shared_ptr<class COsdFontAllocator>	COsdFontAllocatorPtr;

/// Classe de base pour les objets OSD de type texte.
class COsdObjectText : public COsdObject
{
public:
	bool				bRJustif;		//!< \p true si ce texte est justifi  droite
	UINT				nTextSize;		//!< taille du texte
	TCHAR				szText[128];	//!< sauvegarde du texte

	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets
	/// \param[in]	eZOrder	Ordre d'affichage, et cl de recherche pour effacement
	/// \param[in]	bRJustf	\p true si alignement  droite
	COsdObjectText(COsdAllObjects & cOsd, OsdLabelsAndZOrder eZOrder, bool bRJustf) :
		COsdObject(cOsd, eZOrder),
		bRJustif(bRJustf),
		nTextSize(0)
	{
		szText[0] = 0;
	}

	/// Retourne le type de l'objet OSD
	virtual OsdObjType	Type() const {
							return oot_Text; }

	/// Tester si un contenu existe
	virtual bool		HasContents() const {
							return nTextSize > 0; }

	COsdFontAllocatorPtr
						GetFontAllocator() const;

	/// criture d'un nouveau texte dans ce rectangle OSD
	/// \param[in]	pszFmt	Chane format style \p printf
	/// \param[in]	argptr	Liste d'arguments, selon format
	void				SetText(LPCTSTR pszFmt, va_list argptr);

	/// Effacement dfinitif des donnes de cet objet OSD.
	virtual void		Clear();
};

// ====================================================================================
// Dclaration de la classe COsdObjectBmp
// ====================================================================================

class COsdObjectBmp : public COsdObject
{
public:
	HBITMAP		m_hBmp;
	HDC			m_hDC;
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets
	/// \param[in]	eZOrder	Ordre d'affichage, et cl de recherche pour effacement
	/// \param[in]	bRJustf	\p true si alignement  droite
	COsdObjectBmp(COsdAllObjects & cOsd, OsdLabelsAndZOrder eZOrder) :
		COsdObject(cOsd, eZOrder),
		m_hBmp(NULL)
	{
	}

	/// Retourne le type de l'objet OSD
	virtual OsdObjType	Type() const {
							return oot_Bmp; }

	/// Tester si un contenu existe
	virtual bool		HasContents() const {
							return m_hBmp != NULL; }

	virtual ~COsdObjectBmp();
};

// ====================================================================================

typedef shared_ptr<COsdObject>		COsdObjectPtr;

/// Functor de comparaison des pointeurs OSD pour tri sur le Z-Order
struct OsdLess : public binary_function<COsdObjectPtr, COsdObjectPtr, bool>
{
	bool operator()(const COsdObjectPtr & p1, const COsdObjectPtr & p2) const
	{
		return (p1->m_eZOrder < p2->m_eZOrder);
	}
};

typedef set<COsdObjectPtr, OsdLess>	COsdObjectSet;
typedef COsdObjectSet::iterator		COsdObjectIter;

// ====================================================================================
// Dclaration de la classe COsdAllObjects
// ====================================================================================

/// Classe regroupant tous les objets OSD dfinis.
class COsdAllObjects
{
	friend class COsdObjectText;
	friend class COsdLayer;

	DWORD					last_event;		//!< Moment du dernier vnement OSD reu
	RECT					sOsdRect;		//!< Rectangle utilis par la vido, base pour l'OSD
	COsdFontAllocatorPtr	pOsdFontAlloc;	//!< Gestionnaire d'allocation des polices de caractres
	COsdObjectSet			setOsdObjects;
	bool					bEnabled;		//!< tat d'activation de l'OSD

	void		SetOsdObject(COsdObject * pNewObj);

	void		SetOsdText(COsdObjectText * pNewObj, LPCTSTR pszFmt, va_list argptr);

public:
	/// Constructeur
	COsdAllObjects();

	void		SetFontAllocator(COsdFontAllocator * pNewOsdFontAlloc) {
					pOsdFontAlloc.reset(pNewOsdFontAlloc); }

	UINT		GetFontSize() const;

	/// Calcul du rectangle utilis pour la vido ( l'usage de la construction de l'OSD)
	/// \param[in] sRect	Rectangle client de la fentre, calcul par \p CalcCliRect
	/// \return				Rectangle calcul, utilisable pour l'OSD
	static RECT	CalcOsdRect(RECT sRect);

	/// Calculer la hauteur de la police  utiliser
	/// \param[in] sRect Rectangle OSD de la fentre, calcul par \p CalcOsdRect
	static LONG	CalcFontHeight(const RECT & sRect);

	template <typename T>
		void	SetOsdText(LPCTSTR pszFmt, ...) {
					va_list argptr; va_start(argptr, pszFmt);
					SetOsdText(new T(*this), pszFmt, argptr); va_end(argptr); }

	/// Affichage du nom de la chane et du nom de l'mission en cours
	/// \param[in] canal Descripteur de la chane courante
	void	DisplayChannelInfo(const Chaine & canal);

	/// Affichage d'un numro de chane (affichage en cours de saisie)
	/// \param[in] nChanNumber Numro  afficher
	void	DisplayChannelNumber(WORD nChanNumber);

	/// Affichage du volume sonore
	/// \param[in] nCurrentVolume niveau de volume courant
	void	DisplayVolume(long nCurrentVolume);

	/// Affichage du zoom
	/// \param[in] nZoomRatio zoom courant
	void	DisplayZoom(int nZoomRatio);

	/// Affichage du nom d'une piste son
	/// \param[in] canal	Descripteur de la chane courante
	/// \param[in] nTrack	Numro de la piste son
	void	DisplaySoundInfo(const Chaine & canal, int nTrack);

	/// Affichage de l'tat "mute"
	/// \param[in] bMuted true si on est en mode silencieux
	void	DisplayMute(bool bMuted);

	/// Affichage d'un message quelconque (dans le coin infrieur droit)
	void	DisplayBotRightMsg(LPCTSTR pszMsg);

	/// Effacement de tous les objets OSD affichs
	void	Clear();

	/// Mthode  appeler en rponse  l'expiration des dlais lis  l'OSD
	void	ProcessTimeEvent();

	/// Redessiner tous les objets dfinis
	/// \return	\p true si au moins un texte a t raffich
	bool	Redraw(HDC hDC);

	/// Dsactivation de l'affichage OSD : tout message reu pendant la dsactivation
	/// sera mis en attente et affich  la ractivation.
	void	Disable();

	/// Rsactivation de l'affichage OSD, avec affichage effectif des messages qui
	/// sont en attente.
	void	Enable();

	void	BeginChange();
	void	EndChange();

	/// Vrifie si le rectangle de l'OSD a chang. Retourne \p true si oui. Optionnellement,
	/// place le nouveau rectangle dans la variable pointe
	bool	CheckOsdRectChange(LPRECT pNewRect = NULL);

	/// Actualisation des rectangles et de la police d'affichage
	///  utiliser  chaque redimensionnement de la vido
	void	UpdateRectsAndFont();
};

// ====================================================================================
// Dclarations des drivations de la classe COsdObjectText
// ====================================================================================

/// OSD pour Numro de chane et nom de chane
class COsdObjectTopRight : public COsdObjectText
{
	/// Ajustement des dimensions du rectangle pour cet objet
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void	AdjustOsdRect(RECT & sRct) const {
						sRct.bottom = sRct.top + cOSD.GetFontSize();}
public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
	COsdObjectTopRight(COsdAllObjects & cOsd) :
		COsdObjectText(cOsd, osd_Top, true)
	{
	}
};

/// OSD pour Nom de programme (2me ligne)
class COsdObjectTopRightL2 : public COsdObjectText
{
	/// Ajustement des dimensions du rectangle pour cet objet
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void	AdjustOsdRect(RECT & sRct) const {
						UINT nFntHgt = cOSD.GetFontSize();
						sRct.top += nFntHgt;	// Une ligne plus bas
						sRct.bottom = sRct.top + nFntHgt; }
public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
	COsdObjectTopRightL2(COsdAllObjects & cOsd) :
		COsdObjectText(cOsd, osd_Top2, true)
	{
	}
};

/// OSD pour "muet" ou pistes audio
class COsdObjectBottomRight : public COsdObjectText
{
	/// Ajustement des dimensions du rectangle pour cet objet
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void	AdjustOsdRect(RECT & sRct) const {
						sRct.right -= 110;		// Rserver 110 pixels pour l'image audio
						sRct.left += RWdt(sRct) / 3;					// 2/3 largeur ( droite)
						sRct.top = sRct.bottom - cOSD.GetFontSize(); }	// Aligner en bas
public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
	COsdObjectBottomRight(COsdAllObjects & cOsd) :
		COsdObjectText(cOsd, osd_BotRight, true)
	{
	}
};

/// OSD pour Volume et Zoom
class COsdObjectBottomLeft : public COsdObjectText
{
	/// Ajustement des dimensions du rectangle pour cet objet
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void	AdjustOsdRect(RECT & sRct) const {
						sRct.right -= 2*(RWdt(sRct) / 3);				// Tiers de largeur ( gauche)
						sRct.top = sRct.bottom - cOSD.GetFontSize(); }	// Aligner en bas
public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
	COsdObjectBottomLeft(COsdAllObjects & cOsd) :
		COsdObjectText(cOsd, osd_BotLeft, false)
	{
	}
};

class COsdObjectSoundBmp : public COsdObjectBmp
{
	/// Ajustement des dimensions du rectangle pour cet objet
	/// \param[in]	sRct	Le rectangle total utilisable pour l'OSD
	/// \param[out]	sRct	Le rectangle rajust  l'objet courant
	virtual void	AdjustOsdRect(RECT & sRct) const {
						SIZE sSiz = ::GetBmpSize(m_hBmp);
						sRct.left = sRct.right - sSiz.cx;
						sRct.top = sRct.bottom - sSiz.cy; }
public:
	/// Constructeur
	/// \param[in]	cOsd	Rfrence sur l'objet qui contient tous les objets OSD
	COsdObjectSoundBmp(COsdAllObjects & cOsd, UINT nBmpId);

	/// Effacement dfinitif des donnes de cet objet OSD.
	virtual void		Clear() {
							}
};
