/*
 * $Id: record.cpp 278 2011-03-24 21:25:05Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion du contrle des enregistrements (implmentations)
 *
 *	Ce fichier implmente les fonctionnalits ncessaires au dmarrage,
 *	 l'arrt et au contrle des enregistrements, programms ou non.
 **/

#include "stdafx.h"
#include "record.h"
#include "channels.h"
#include "grabber.h"
#include "capture.h"

/**
 * Table des descripteurs d'enregistrements en cours de traitement
 **/
Enregistrements	enregistrements_actuels;

/**
 * Dmarrage d'un enregistrement.
 * \param[in]	pCapt		Pointeur sur un objet driv de \p Mpeg_TS_Input, implmentant les
 *							mthode de traitement des chantillons  enregistrer
 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer, ou bien
 *							\p STREAM_PSEUDO_INDEX pour spcifier tout le multiplex
 * \param[in]	eMeth		Mthode d'enregistrement
 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
 *							y sont rcuprs
 * \return	\p true si succs, \p false si chec
 **/
bool Enregistrement::start(Mpeg_TS_InpPtr pCapt, int ixChaine, 
						   MethodeEnregistrement eMeth, Programme * pProg)
{
	UINT32 cleCh = indexToCle(ixChaine);

	stop(epr_interrompu);	// prcaution

	if (!pGraph)
		return false;
	nErrCode = 0;

	GrabberData & sGrabber = pGraph->GetGrabberInterface();

	if (sGrabber.pInterface && cleCh!=0 && sGrabber.pCallback) {
		// Nombre d'enregistrements avant le dmarrage de celui-ci :
		int nPrevRecCount = enregistrements_actuels.count_recording();

		if (pProg) {
			memcpy(this, pProg, sizeof(RecordInfo));
			pProg->etat = epr_encours;
			programmation_modifiee = true;
		}

		CSampleGrabber & cSg = dynamic_cast<CSampleGrabber &>(*sGrabber.pCallback);

		if (cSg.cMutex.Lock()) {	// On a l'accs au mutex
			pCapture = pCapt;
			cSg.cMutex.Unlock();
			cleChaine = cleCh;
			methode = eMeth;

			// S'il n'y avait aucun enregitrement en cours auparavant, on dmarre le "callback" maintenant :
			if (nPrevRecCount==0) {
				sGrabber.SetCallback();
				cSg.StartThread();	// Dmarrage du "thread"
			}
			return true;
		}
	}

	// Si on arrive ici, c'est que la requte a chou
	return false;
}

/**
 * Prise en compte d'une programmation affectant une chane dj en cours d'enregistrement.
 * Cette fonction est appele dans le cas o un enregistrement programm doit dmarrer, et
 * qu'il est constat que la chane correspondante a dj un enregistrement en cours.
 * \param[in]	pProg	Pointeur sur la programmation horaire ayant servi  dmarrer
 *						l'enregistrement
 * \retval		\p true si succs
 * \retval		\p false si une incompatibilit est rencontre : la programmation demande
 *				une mthode d'enregistrement ou une piste audio qui diffrent de celles de
 *				l'enregistrement en cours ; dans ce cas, l'enregistrement est mis en erreur
 **/
bool Enregistrement::modify_prog(Programme * pProg)
{
	if (pProg && DiffTime(pProg->fin, fin)>0) {
		programmation_modifiee = true;
		if (pProg->methode!=methode || pProg->audio!=audio) {
			// Incompatibilit : la programmation demande une mthode d'enregistrement ou une
			// piste audio qui diffrent de celles de l'enregistrement en cours
			pProg->etat = epr_erreur;					// tat de la nouvelle programmation utilise
			return false;
		}

		int ixProg = trouve_prog_par_nom(nom);

		if (ixProg>=0) {
			Programmes[ixProg].etat = epr_supplante;	// tat de l'ancienne programmation utilise
			record_list_updated();
		}

		fin = pProg->fin;
		apres = pProg->apres;
		pProg->etat = epr_encours;						// tat de la nouvelle programmation utilise
		strcpy_T(nom, pProg->nom);
	}
	return true;
}

/**
 * Arrt de l'enregistrement.
 * \param[in]	raison	Raison de l'arrt (tat de programmation suivant)
 * \retval	Action  effectuer ensuite
 * \retval	\p apr_null si l'enregistrement n'tait pas en cours.
 **/
ApresEnregistrement Enregistrement::stop(EtatProgramme raison)
{
	ApresEnregistrement after = apr_null;

	if (pGraph && recording()) {
		myprintf(BOLD(TEXT("Arrt d'enregistrement")) EOL);
		after = apres;

		if (cleChaine!=0) {
			if (raison!=epr_null) {
				int ixProg = trouve_prog_par_nom(nom);

				if (ixProg>=0) {
					Programmes[ixProg].etat = raison;
					programmation_modifiee = true;
				}
			}

			GrabberData & sGrabber = pGraph->GetGrabberInterface();

			if (pCapture && sGrabber.pCallback) {
				CSampleGrabber & cSg = dynamic_cast<CSampleGrabber &>(*sGrabber.pCallback);

				if (cSg.cMutex.Lock()) {
					pCapture.reset();
					cSg.cMutex.Unlock();
					cleChaine = 0;	// Ncessaire pour le comptage

					// Nombre d'enregistrement qui restent en cours :
					int nPrevRecCount = enregistrements_actuels.count_recording();

					// S'il n'y a plus aucun enregistrement en cours, on arrte le "callback" maintenant :
					if (nPrevRecCount==0) {
						sGrabber.ResetCallback();
						cSg.StopThread();
					}
				}
			}
		}
		nom[0] = 0;
		cleChaine = 0;		// Parfois redondant, mais tant pis
		methode = meth_TS;	// Surtout pour qu'il ne soit plus spcifi "multiplex" le cas chant
		apres = apr_null;
	}
	return after;
}

/**
 * Obtenir le nom de la chane enregistre (retourne un "*" s'il s'agit du multiplex entier) :
 * \retval	Pointeur sur le nom de la chane
 * \retval	"*"		S'il s'agit du multiplex entier
 * \retval	"??"	(erreur interne)
 **/
LPCSTR Enregistrement::channelName() const
{
	if (cleChaine==STREAM_PSEUDO_CLE)
		return "*";

	int ixChaine = Canaux.trouve_par_cle(cleChaine);

	if (ixChaine_ok(ixChaine))
		return Canaux[ixChaine].nom;
	return "??";
}

/// Obtenir le nom de la chane enregistre, pour affichage
tstring Enregistrement::dispChannelName() const
{
	TCHAR	szWrk[64] = TEXT("");

	if (recording()) {
		strcpy_T(szWrk, channelName());
		if (*szWrk == TCHAR('*'))
			strcpy_T(szWrk, TEXT("[ multiplex ]"));
	}
	return szWrk;
}

/**
 * Calcul du temps restant avant la fin de l'enregistrement.
 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
 * \retval	Temps restant, en millisecondes
 **/
long Enregistrement::temps_restant(const SYSTEMTIME & localtime) const
{
	if (!recording() || apres==apr_null)
		// Si pas en train d'enregistrer, ou bien si pas d'arrt programm
		return DUREE_INFINIE;

	long reste = DiffTime(fin, localtime);

	return max(reste, 0);
}

	/// Compter les enregistrements en cours
int Enregistrements::count_recording() const
{
	int nRes = 0;

	for (int nInx=0; nInx < count(); ++nInx)
		if (cRecord[nInx].recording())
			++nRes;
	return nRes;
}

/**
 * Dtermine si on enregistre actuellement
 *
 * \retval	Cl de la premire chane enregistre trouve
 * \retval	0 si aucun enregistrement n'est en cours.
 **/
UINT32 Enregistrements::recording() const
{
	for (int nInx=0; nInx < count(); nInx++) {
		const Enregistrement & enreg = cRecord[nInx];

		if (enreg.recording())
			return enreg.cleChaine;
	}
	return 0;
}

/**
 * Arrte tous les enregistrements
 **/
void Enregistrements::stop_all()
{
	for (int nInx=0; nInx < count(); nInx++)
		cRecord[nInx].stop(epr_null);	// epr_null -> ne pas mmoriser cet arrt
}

/**
 * Rechercher si une chane est en cours d'enregistrement,
 * et si oui, obtenir l'index du descripteur de cet enregistrement.
 * \param[in]	cleChaine	Cl de la chane concerne
 * \retval		Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval		-1 si la chane n'est pas en cours d'enregistrement.
 **/
int Enregistrements::getRecordingDescrByKey(UINT32 cleChaine) const
{
	if (cleChaine!=0) {
		for (int nInx=0; nInx < count(); nInx++) {
			if (cRecord[nInx].isChannel(cleChaine))
				return nInx;
		}
	}
	return -1;
}

/**
 * Rechercher si un multiplex est en cours d'enregistrement,
 * et si oui, obtenir l'index du descripteur de cet enregistrement.
 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval	-1 si aucun multiplex n'est en cours d'enregistrement.
 **/
int Enregistrements::getRecordingStream() const
{
	for (int nInx=0; nInx < count(); nInx++) {
		const Enregistrement cRcrd = cRecord[nInx];

		if (cRcrd.cleChaine!=0 && cRcrd.methode == meth_multiplex)
			return nInx;
	}
	return -1;
}

/**
 * Chercher un descripteur d'enregistrement libre
 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval	-1 si tous les descripteurs possibles sont utiliss.
 **/
int Enregistrements::getFreeDescriptor() const
{
	for (int nInx=0; nInx < count(); nInx++) {
		if (!cRecord[nInx].recording())
			return nInx;
	}
	myprintf(TEXT("Plus aucun descripteur d'enregistrement disponible") EOL);
	return -1;
}

/**
 * Trouver l'index d'un enregistrement effectif et actif sur la base de son nom.
 *
 * \param[in]	nom	Nom de l'enregistrement recherch
 * \retval	 Index numrique de l'enregistrement
 * \retval	 -1 si rien trouv.
 **/
int Enregistrements::trouve_enreg_par_nom(LPCTSTR nom) const
{
	for (int nInx=0; nInx < count(); nInx++) {
		const Enregistrement & enreg = cRecord[nInx];

		if (enreg.recording() && _tcscmp(enreg.nom, nom)==0)
			return nInx;
	}

	return -1;
}

/**
 * Calcul du temps restant avant la prochaine fin d'enregistrement, parmi
 * tous les enregistrements en cours.
 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
 * \retval	Temps restant, en millisecondes.
 * \retval	\p INFINITE si aucun enregistrement n'est en cours, ou bien si
 *			aucun de ceux-ci n'a d'chance programme.
 **/
long Enregistrements::temps_restant(const SYSTEMTIME & localtime) const
{
	long nRes = DUREE_INFINIE;
	long nTmp;

	for (int nInx = 0; nInx < count(); nInx++) {
		nTmp = cRecord[nInx].temps_restant(localtime);
		nRes = min(nRes, nTmp);
	}
	return nRes;
}

/**
 * Obtenir le pointeur sur le nom d'une programmation si celle-ci a t explicitement
 * nomme, et NULL dans le cas contraire :
 **/
static LPTSTR getProgName(Programme * pProg)
{
	if (pProg!=NULL && !pProg->nomAuto && *pProg->nom!=0)
		return pProg->nom;
	return NULL;
}

/**
 * Dmarrage d'un enregistrement en mode TS.
 * Le tuner est suppos tre dj syntonis sur la bonne frquence.
 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer
 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
 *							y sont rcuprs
 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
 **/
int Enregistrements::start_record_ts(int ixChaine, Programme * pProg)
{
	if (!ixChaine_ok(ixChaine))
		return -1;

	int					ixRecDescr	= getFreeDescriptor();

	if (ixRecDescr < 0)
		return -1;

	const Chaine &		canal		= Canaux[ixChaine];
	NomFichierAvecMotif	nom_fichier(video_dir, getProgName(pProg), canal.nom, canal.nNumeroChaine, TEXT("ts"));
	HANDLE				hFile		= nom_fichier.Create();

	if (hFile == INVALID_HANDLE_VALUE) {
		myprintf(ERRX(TEXT("Erreur cration '%s' code %u")), nom_fichier.nom, GetLastError());
		return -1;
	}

	Mpeg_TS_InpPtr	pCapture(new CCapture_TS(hFile, canal));

	if (pCapture==NULL)
		return -1;

	myprintf(BOLD(TEXT("Enregistre TS")) TEXT(", video=%i, pmt=%i") EOL TEXT("\"%s\"") EOL,
		canal.video_pid, canal.pmt_pid, nom_fichier.nom);
	cRecord[ixRecDescr].start(pCapture, ixChaine, meth_TS, pProg);
	return ixRecDescr;
}

/**
 * Dmarrage d'un enregistrement en mode PS.
 * Le tuner est suppos tre dj syntonis sur la bonne frquence.
 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer
 * \param[in]	ixSon		Index de la piste audio dans la table \p sons de cette chane
 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
 *							y sont rcuprs
 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
 **/
int Enregistrements::start_record_ps(int ixChaine, int ixSon, Programme * pProg)
{
	if (!ixChaine_ok(ixChaine))
		return -1;

	int					ixRecDescr	= getFreeDescriptor();

	const Chaine &		canal		= Canaux[ixChaine];

	if (ixRecDescr < 0 || ixSon < 0 || ixSon >= canal.nbr)
		return -1;

	const une_piste &	sAudio		= canal.tbl[ixSon];

	if (!canal.isPsRecordAvailable())
		return -1;		// Enregistrement PS non disponible

	NomFichierAvecMotif	nom_fichier(video_dir, getProgName(pProg), canal.nom, canal.nNumeroChaine, TEXT("mpg"));
	UINT16				video_pid	= canal.video_pid;
	UINT16				audio_pid	= sAudio.pid;
	HANDLE				hFile		= nom_fichier.Create();

	if (hFile == INVALID_HANDLE_VALUE) {
		myprintf(TEXT("Erreur cration '%s' code %u") EOL, nom_fichier.nom, GetLastError());
		return -1;
	}

	Mpeg_TS_InpPtr	pCapture(new CCapture_PS(hFile, video_pid, audio_pid));

	if (!pCapture.get())
		return -1;

	myprintf(BOLD(TEXT("Enregistre PS")) TEXT(", video=%i, audio=%i") EOL TEXT("\"%s\"") EOL,
		video_pid, audio_pid, nom_fichier.nom);
	cRecord[ixRecDescr].start(pCapture, ixChaine, meth_PS, pProg);
	return ixRecDescr;
}

/**
 * Dmarrage d'un enregistrement global du multiplex.
 * \param[in]	pProg	Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
 *						l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
 *						y sont rcuprs
 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
 **/
int Enregistrements::start_record_stream(Programme * pProg)
{
	if (getRecordingStream() > 0)
		return -1;		// On n'enregistre pas deux fois simultanment le mme multiplex

	int					ixRecDescr	= getFreeDescriptor();

	if (ixRecDescr < 0)
		return -1;
	CHAR				szMxName[128] = "Multiplex";	// Par dfaut

	// Utilisation du nom du multiplex si disponible
	if (ixChaine_ok(ixChaineCourante)) {
		Chaine & sChan = Canaux[ixChaineCourante];

		if (*sChan.groupe)
			sprintf_s(szMxName, "Multiplex %s", sChan.groupe);
	}

	NomFichierAvecMotif	nom_fichier(video_dir, getProgName(pProg), szMxName, -1, TEXT("ts"));
	HANDLE				hFile		= nom_fichier.Create();

	if (hFile == INVALID_HANDLE_VALUE) {
		myprintf(TEXT("Erreur cration '%s' code %u") EOL, nom_fichier.nom, GetLastError());
		return -1;
	}

	Mpeg_TS_InpPtr	pCapture(new CCapture_Stream(hFile));

	if (!pCapture.get())
		return -1;

	myprintf(BOLD(TEXT("Enregistre multiplex")) EOL TEXT("\"%s\"") EOL, nom_fichier.nom);
	cRecord[ixRecDescr].start(pCapture, STREAM_PSEUDO_INDEX, meth_multiplex, pProg);
	return ixRecDescr;
}


#ifdef _DIAGNOSTIC
	/**
	 * Dmarrage d'un enregistrement global d'un multiplex inconnu.
	 * \param[in]	nFreq	Frquence  slectionner au pralable
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
	 * \note	Utilis seulement en diagnostic, pour obtenir des informations sur un multiplex
	 *			apparemment existant, mais dans lequel le programme ne peut trouver aucune chane.
	 **/
	int Enregistrements::start_record_undefstream(int nFreq)
	{
		zappe_index(-1);					// Dsactiver la chane courante
		change_frequence(nFreq, DVBT_BANDWIDTH);
		// Sleep(sTunProps.delai_tuner1);
		return start_record_stream();
	}
#endif
