/*
 * $Id: record.h 306 2011-09-28 09:28:04Z gingko $
 * Copyright (C) 2007 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion du contrle des enregistrements (dfinitions)
 *
 *	Ce fichier contient les dfinitions ncessaires au dmarrage,
 *	 l'arrt et au contrle des enregistrements, programms ou non.
 **/

#pragma once

#include "recprog.h"
#include "capture.h"

/// Nombre max d'enregistrements simultans sur un mme multiplex
#define NB_MAX_ENREG	4

/**
 * Dure maximale qui spare deux appels au timer de dbut d'enregistrement.
 * Si le temps avant l'enregistrement suivant est plus long que ce dlai,
 * le timer est programm sur cette valeur, et est donc de fait recalcul
 *  intervalles identiques  cette valeur jusqu' ce que la dure restante
 * soit infrieure.
 **/
#if _DEBUG
  #define DUREE_RECALCUL_TIMERS 10			// Valeur en secondes
#else
  #define DUREE_RECALCUL_TIMERS (10*60)		// Valeur en secondes
#endif

#define STREAM_PSEUDO_SID 0xffff		//!< Pseudo SID pour l'enregistrement du multiplex \deprecated Voir #indexToSID
#define STREAM_PSEUDO_INDEX -2			//!< Pseudo index de chane pour l'enregistrement du multiplex

/// Description d'un enregistrement (capture) de chane qui est en cours de traitement
class Enregistrement : public RecordInfo {
public:
	Mpeg_TS_InpPtr	pCapture;
	DWORD			nErrCode;		//!< Code d'erreur ventuel

	/// Constructeur
	Enregistrement() :
		nErrCode(0)
	{
	}

	/// Dterminer si cet enregistrement de chane est couramment actif
	bool recording() const {
			return cleChaine != 0;}

	/// Dterminer si cet enregistrement de chane correspond  la cl \p cleCh, donne en paramtre
	bool isChannel(UINT32 cleCh) const {
		if (cleCh == STREAM_PSEUDO_CLE)
			return methode == meth_multiplex;
		else
			return cleChaine ==  cleCh; }

	/**
	 * Dmarrage d'un enregistrement.
	 * \param[in]	pCapt		Pointeur sur un objet driv de \p Mpeg_TS_Input, implmentant les
	 *							mthode de traitement des chantillons  enregistrer
	 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer, ou bien
	 *							\p STREAM_PSEUDO_INDEX pour spcifier tout le multiplex
	 * \param[in]	eMeth		Mthode d'enregistrement
	 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
	 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
	 *							y sont rcuprs
	 * \return	\p true si succs, \p false si chec
	 **/
	bool start(Mpeg_TS_InpPtr pCapt, int ixChaine,
			   MethodeEnregistrement eMeth, Programme * pProg);

	/**
	 * Prise en compte d'une programmation affectant une chane dj en cours d'enregistrement.
	 * Cette fonction est appele dans le cas o un enregistrement programm doit dmarrer, et
	 * qu'il est constat que la chane correspondante a dj un enregistrement en cours.
	 * \param[in]	pProg	Pointeur sur la programmation horaire ayant servi  dmarrer
	 *						l'enregistrement
	 * \retval		\p true si succs
	 * \retval		\p false si une incompatibilit est rencontre : la programmation demande
	 *				une mthode d'enregistrement ou une piste audio qui diffrent de celles de
	 *				l'enregistrement en cours ; dans ce cas, l'enregistrement est mis en erreur
	 * \todo Faire en sorte qu'une incompatibilit ne provoque pas d'erreur, mais  la place,
	 * amne l'arrt de l'enregistrement en cours suivi immdiatement du dmarrage d'un nouvel
	 * enregistrement utilisant les nouveaux paramtres
	 **/
	bool modify_prog(Programme * pProg);

	/**
	 * Calcul du temps restant avant la fin de l'enregistrement.
	 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
	 * \retval	Temps restant, en millisecondes
	 **/
	long temps_restant(const SYSTEMTIME & localtime) const;

	/**
	 * Arrt de l'enregistrement.
	 * \param[in]	raison	Raison de l'arrt (tat de programmation suivant)
	 * \retval	Action  effectuer ensuite
	 * \retval	\p apr_null si l'enregistrement n'tait pas en cours.
	 **/
	ApresEnregistrement stop(EtatProgramme raison);

	/**
	 * Obtenir le nom de la chane enregistre (retourne un "*" s'il s'agit du multiplex entier) :
	 * \retval	Pointeur sur le nom de la chane
	 * \retval	"*"		S'il s'agit du multiplex entier
	 * \retval	"??"	(erreur interne)
	 **/
	LPCSTR channelName() const;

	/// Obtenir le nom de la chane enregistre, pour affichage
	tstring dispChannelName() const;
};

class Enregistrements
{
	Enregistrement cRecord[NB_MAX_ENREG];
public:
	Enregistrement &	operator [] (int nSub) {
								return cRecord[nSub];}

	const Enregistrement &
						operator [] (int nSub) const {
							return cRecord[nSub];}

	int					count() const {
							return _countof(cRecord);}

	/// Compter les enregistrements en cours
	int					count_recording() const;

	/**
	 * Dtermine si on enregistre actuellement
	 *
	 * \retval	Cl de la premire chane enregistre trouve
	 * \retval	0 si aucun enregistrement n'est en cours.
	 **/
	UINT32				recording() const;

	/**
	 * Arrt de tous les enregistrements en cours
	 **/
	void				stop_all();

	/**
	 * Rechercher si une chane est en cours d'enregistrement,
	 * et si oui, obtenir l'index du descripteur de cet enregistrement.
	 * \param[in]	cleChaine	Cl de la chane concerne
	 * \retval		Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval		-1 si la chane n'est pas en cours d'enregistrement.
	 **/
	int					getRecordingDescrByKey(UINT32 cleChaine) const;

	/**
	 * Rechercher si une chane est en cours d'enregistrement,
	 * et si oui, obtenir l'index du descripteur de cet enregistrement.
	 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane concerne, ou bien
	 *							\p STREAM_PSEUDO_INDEX pour spcifier tout le multiplex courant
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si la chane n'est pas en cours d'enregistrement.
	 **/
	int					getRecordingDescr(int ixChaine) const {
							return getRecordingDescrByKey(indexToCle(ixChaine));}

	/**
	 * Rechercher si un multiplex est en cours d'enregistrement,
	 * et si oui, obtenir l'index du descripteur de cet enregistrement.
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun multiplex n'est en cours d'enregistrement.
	 **/
	int					getRecordingStream() const;

	/**
	 * Chercher un descripteur d'enregistrement libre
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si tous les descripteurs possibles sont utiliss.
	 **/
	int					getFreeDescriptor() const;

	/**
	 * Trouver l'index d'un enregistrement effectif et actif sur la base de son nom.
	 *
	 * \param[in]	nom	Nom de l'enregistrement recherch
	 * \retval	 Index numrique de l'enregistrement
	 * \retval	 -1 si rien trouv.
	 **/
	int					trouve_enreg_par_nom(LPCTSTR nom) const;

	/**
	 * Calcul du temps restant avant la prochaine fin d'enregistrement, parmi
	 * tous les enregistrements en cours.
	 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
	 * \retval	Temps restant, en millisecondes.
	 * \retval	\p INFINITE si aucun enregistrement n'est en cours, ou bien si
	 *			aucun de ceux-ci n'a d'chance programme.
	 **/
	long				temps_restant(const SYSTEMTIME & localtime) const;

	/**
	 * Dmarrage d'un enregistrement en mode TS.
	 * Le tuner est suppos tre dj syntonis sur la bonne frquence.
	 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer
	 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
	 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
	 *							y sont rcuprs
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
	 * \todo Retirer la restriction sur la syntonisation, et ajouter le paramtre \p ixSon
	 * pour choisir de n'enregistrer qu'une seule piste audio, si ce paramtre n'est pas -1.
	 **/
	int					start_record_ts(int ixChaine, Programme * pProg=NULL);

	/**
	 * Dmarrage d'un enregistrement en mode PS.
	 * Le tuner est suppos tre dj syntonis sur la bonne frquence.
	 * \param[in]	ixChaine	Index (dans la table #Canaux) de la chane  enregistrer
	 * \param[in]	ixSon		Index de la piste audio dans la table \p sons de cette chane
	 * \param[in]	pProg		Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
	 *							l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
	 *							y sont rcuprs
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
	 * \todo Retirer la restriction sur la syntonisation, et faire en sorte que toutes
	 * les pistes son soient enregistres si \p ixSon = -1
	 **/
	int					start_record_ps(int ixChaine, int ixSon, Programme * pProg=NULL);

	/**
	 * Dmarrage d'un enregistrement global du multiplex.
	 * \param[in]	pProg	Pointeur facultatif sur la programmation horaire ayant servi  dmarrer
	 *						l'enregistrement : s'il n'est pas \p NULL, des paramtres additionnels
	 *						y sont rcuprs
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
	 * \todo Ajouter un paramtre (par ex. \p ixChaine) pour que soit pralablement syntonis
	 * le multiplex correspondant  la chane dsigne.
	 **/
	int					start_record_stream(Programme * pProg=NULL);

#ifdef _DIAGNOSTIC
	/**
	 * Dmarrage d'un enregistrement global d'un multiplex inconnu.
	 * \param[in]	nFreq	Frquence  slectionner au pralable
	 * \retval	Index du descripteur dans la table des descripteurs d'enregistrements en cours
	 * \retval	-1 si aucun enregistrement n'a pu dmarrer.
	 * \note	Utilis seulement en diagnostic, pour obtenir des informations sur un multiplex
	 *			apparemment existant, mais dans lequel le programme ne peut trouver aucune chane.
	 **/
	int					start_record_undefstream(int nFreq);
#endif
};

/**
 * Table des enregistrements d'missions en cours de traitement
 * (parfois dnomms "grabbers" en raison de l'usage d'un filtre
 * "Sample grabber" pour chacun de ces enregistrements)
 **/
extern Enregistrements	enregistrements_actuels;
