/*
 * $Id: recprog.h 385 2020-09-25 12:21:44Z gingko $
 * Copyright (C) 2007 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion des enregistrements programms (dfinitions)
 *
 *	Ce fichier contient les dfinitions de classes, numrations et fonctions
 *	ncessaires  la programmation des enregistrements.
 **/

#pragma once

#include "dlgutils.h"
#include "resource.h"

/// Constante reprsentant une dure d'une seconde
#define DUREE_1SEC	1000L				// Valeur en millisecondes

/// Constante reprsentant une dure d'une minute
#define DUREE_1MN	(60 * DUREE_1SEC)	// Valeur en millisecondes

/// Constante reprsentant une dure d'une heure
#define DUREE_1H	(60 * DUREE_1MN)	// Valeur en millisecondes

/// Constante reprsentant une dure de 24h
#define	DUREE_24H	(24 * DUREE_1H)		// Valeur en millisecondes

/// Temps de dcalage pour l'ajout au planificateur de tches
#define			DUREE_PLANIF				DUREE_1MN

/// Indique quoi faire aprs un enregistrement
enum ApresEnregistrement {
	apr_null = -1,	//!< -1: Utilis dans les enregistrements en cours, pour indiquer
					//!<	 qu'aucun horaire de fin n'est associ  un enregistrement
	apr_rien = 0,	//!< 0: On ne fait rien
	apr_quitter,	//!< 1: On quitte l'application
	apr_eteindrePC,	//!< 2: On quitte l'application et on teint l'ordinateur
	apr_eteindrePF,	//!< 3: On quitte l'application et on teint l'ordinateur (forc)
	apr_veille,		//!< 4: On met l'ordinateur en veille (sans quitter)
	apr_veille_pr	//!< 5: On met l'ordinateur en veille prolonge (sans quitter)
};

/// Statut d'un enregistrement par rapport au planificateur de tches
enum PlanificateurTache {
	planif_sans,	//!< 0: non planifie
	planif_avec,	//!< 1: planifie (ajout effectu)
	planif_ajout	//!< 2: planifie ( ajouter au planificateur lors de la validation)
};

/// tat d'activation d'un enregistrement programm
enum EtatProgramme {
	epr_actif,		//!< L'enregistrement est actif (programm et  venir)
	epr_encours,	//!< L'enregistrement en cours
	epr_inactif,	//!< Dsactiv par l'utilisateur
	epr_interrompu,	//!< Manuellement arrt aprs dmarrage
	epr_supplante,	//!< Heure de fin modifie aprs dmarrage
	epr_termine,	//!< Heure de fin atteinte
	epr_erreur,		//!< Une erreur s'est produite  la programmation
	epr_fileerr,	//!< Erreur du systme de fichiers  l'enregistrement
	epr_null = -1	//!< Pseudo-valeur pour dire "ne pas modifier"
};

typedef TCHAR NomProg[64];	//!< Type pour nom donn  une programmation

/// Timer de rveil
class CWaitTimer
{
	HANDLE	hWTimer;
	LONG	nLastDelay;
public:
	CWaitTimer() :
		hWTimer(NULL),
		nLastDelay(DUREE_INFINIE)
	{
	}

	/** \brief Dfinir  un temps donn
	 * \param[in]	nDelay	Temps avant rveil (millisecondes d'attente)
	 * \return	\p true ou \p false selon succs ou chec
	 **/
	bool Set(UINT nDelay);

	/// Annulation du timer de rveil
	void Cancel();

	/// Obtenir le temps d'attente courant
	LONG GetCurrent() const {
			return nLastDelay; }

	/// Destructeur
	~CWaitTimer()
	{
		if (hWTimer)
			CloseHandle(hWTimer);
	}
};

/// Structure commune aux programmations d'enregistrement et aux enregistrements en cours
class RecordInfo
{
public:
	NomProg					nom;			//!< Nom de la programmation si applicable
	UINT32					cleChaine;		//!< cl de la chane enregistre
	WORD					sidChaine;		//!< \deprecated SID de la chane enregistre
	WORD					nNumeroChaine;	//!< Numro de la chane
	ApresEnregistrement		apres;			//!< Quoi faire aprs la fin de l'enregistrement
	MethodeEnregistrement	methode;		//!< Mthode d'enregistrement (TS, PS, Multiplex)
	AudioMode				audio;			//!< Mode audio utilis /  utiliser
	bool					reveil;			//!< Sortir du mode veille pour enregistrer
	SYSTEMTIME				fin;			//!< Date & heure de fin

	// Mini-constructeur :
	RecordInfo() {
		memset(this, 0, sizeof(*this));
		apres = apr_null; }
};

/**
 * Formatage d'une heure de dbut, pour affichage
 * \param[in]	sTime	Heure  formater
 * \param[in]	pszBuf	Tampon recevant la chane formate
 * \param[in]	nCount	Taille du tampon
 * \retval		Nombre de caractres crits dans le tampon.
 **/
UINT32	formate_heure_debut(const SYSTEMTIME & sTime, LPTSTR pszBuf, UINT32 nCount);

/// Dfinition d'une programmation d'enregistrement
class Programme : public RecordInfo
{
	static const LPCTSTR t_states[];
public:
	SYSTEMTIME				debut;		//!< Date & heure de dbut

	/**
	 * \brief Registre de bits indiquant les jours de rptition des enregistrements
	 *
	 * (bit 0: Dimanche, bit 1: Lundi, ...., bit 6: Samedi)
	 **/
	BYTE					repetition;

	PlanificateurTache		tache;		//!< Statut au regard du planificateur de tches
	EtatProgramme			etat;		//!< tat (d'activation) de la programmation
	bool					nomAuto;	//!< \p true si nom gnr automatiquement

	/// Constructeur
	Programme();

#ifdef _XMLUTILS_H_INCLUDED_

	/// numration corresopondant aux tags XML utiliss pour la sauvegarde d'une programmation
	enum TagListProg {
		tlp_Nom,
		tlp_NomAuto,
		tlp_Chaine,
		tlp_ServiceID,
		tlp_Cle,
		tlp_Debut,
		tlp_Fin,
		tlp_Methode,
		tlp_Audio,
		tlp_Reveil,
		tlp_After,
		tlp_Tache,
		tlp_Etat,
		tlp_Repetitions,
		tlp_S_Annee,	// compatibilit anciennes versions
		tlp_E_Annee		// compatibilit anciennes versions
	};

	/// numration corresopondant aux tags XML utiliss pour la sauvegarde de l'heure et de la date
	enum TagListTime {
		tlt_Annee,
		tlt_Mois,
		tlt_Jour_Semaine,
		tlt_Jour,
		tlt_Heure,
		tlt_Minute,
		tlt_Seconde
	};

	/// Noms des tags XML utiliss pour la sauvegarde d'une programmation
	static LPCSTR pszTagStrList;

	/// Noms des tags XML utiliss pour la sauvegarde de l'heure et de la date
	static LPCSTR pszTimeTagStrList;

	/**
	 * Extraire la date et l'heure d'un noeud XML
	 * \param[in]	pXmlElm		lment (parent) des tags recherchs
	 * \param[in]	pszTagList	List de tags  utiliser
	 * \return		Date et heure retourns
	 **/
	static SYSTEMTIME XmlToTime(const TiXmlElement * pXmlElm, LPCSTR pszTagList);

	/**
	 * Gnrer le noeud XML correspondant  une structure date et heure
	 * \param[in]	sTime	Date et heure  inclure
	 * \param[in]	pszTag	Tag du noeud parent
	 * \return		Noeud XML gnr
	 **/
	static TiXmlElement * TimeToXml(SYSTEMTIME sTime, LPCSTR pszTag);

	/**
	 * Gnrer le noeud XML correspondant  variable de rptition hebdomadaire
	 * \param[in]	nRep	Flags contenant les jours de rptition
	 * \param[in]	pszTag	Tag du noeud parent
	 * \return		Noeud XML gnr
	 **/
	static TiXmlElement * RepeatToXml(BYTE nRep, LPCSTR pszTag);

	/**
	 * Extraire les informations de rptitions d'un noeud XML
	 * \param[in]	pXmlElm		lment (parent) des tags recherchs
	 **/
	static BYTE XmlToRepeat(const TiXmlElement * pXmlElm);

	/**
	 * Chargement d'une programmation  partir d'un noeud TinyXML construit
	 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	bool SetFromXMLElement(const TiXmlElement * pXmlElm);

	/**
	 * Construction d'un noeud TinyXML reprsentant la dfinition de la programmation
	 * \return	Noeud XML construit
	 **/
	TiXmlElement * GetAsXMLElement() const;
#endif

	/**
	 * \brief Vrification de la validit de la programmation
	 *
	 * \param[in]	prog_modif	Si non NULL, pointeur sur la programmation qu'on modifie (pour
	 *							viter de la comparer  elle-mme lors des tests de recouvrement)
	 * \retval		\p true ou \p false selon succs ou chec, respectivement
	 **/
	bool	verifie(const Programme * prog_modif=NULL) const;

	/**
	 * Affichage du message d'avertissement en cas de conflit dans une programmation
	 * \param[in]	pszMessage	Texte du message  afficher
	 * \return		\p true ou \p false selon que l'utilisateur souhaite nanmoins poursuivre ou pas
	 **/
	static bool
			warn_prog(LPCTSTR pszMessage);

	/**
	 * \brief Calcul du temps restant avant dbut enregistrement en millisecondes.
	 *
	 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
	 **/
	long	temps_restant(const SYSTEMTIME & localtime) const;

	/**
	 * \brief Ajustement des donnes si rptitions programmes et horaire de fin dpass
	 *
	 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
	 * \retval		\p true si un ajustement a eu lieu
	 **/
	bool	ajustementDesRepetitions(const SYSTEMTIME & localtime);

#if USE_TASKMAN_PASSWORD
	/**
	 * \brief Ajout de cette programmation aux tches programmes
	 *
	 * \param[in]	username	Nom d'utilisateur
	 * \param[in]	motdepasse	Mot de passe
	 * \retval		\p true ou \p false selon succs ou chec, respectivement
	 **/
	bool	ajouterTacheProgrammee(LPCWSTR username, LPCWSTR motdepasse) const;
#else
	/**
	 * \brief Ajout de cette programmation aux tches programmes
	 *
	 * \retval		\p true ou \p false selon succs ou chec, respectivement
	 **/
	bool	ajouterTacheProgrammee() const;
#endif

	/**
	 * \brief Suppression de cette programmation de la liste des tches programmes
	 *
	 * \retval		\p true ou \p false selon succs ou chec, respectivement
	 **/
	bool	supprimerTacheProgrammee() const;

	/**
	 * \brief Gnration du nom de tche programme correspondant
	 *
	 * \param[in]	pstr	Pointeur sur le tampon qui va recevoir le nom
	 * \param[in]	bufSize	Taille de ce tampon
	 **/
	void	genereNomTache(LPWSTR pstr, UINT bufSize) const;

	/// Retourne \p true si la programmation est couramment active
	bool	active() const {
				return etat==epr_actif || etat==epr_encours;}

	/// Retourne \p true si intersection temporelle avec la programmation \p p2
	bool	overlap(const Programme & p2) const;

	/// Retourne \p true si on a affaire  une programmation d'enregistrement de multiplex
	bool	isMultiplexProg() const {
				return methode==meth_multiplex;}

	/**
	 * \brief Comparaison en vue de tri sur l'heure de dbut
	 *
	 * Utilis par la fonction std::sort
	 * \retval \p true si cette programmation se classe avant la programmation \p p2 dans le temps.
	 **/
	bool	operator < (const Programme & p2) const;

	/**
	 * Formatage de l'heure de dbut, pour affichage
	 * \param[in]	pszBuf	Tampon recevant la chane formate
	 * \param[in]	nCount	Taille du tampon
	 * \retval		Nombre de caractres crits dans le tampon.
	 **/
	UINT32	formate_heure_debut(LPTSTR pszBuf, UINT32 nCount) const {
				return ::formate_heure_debut(debut, pszBuf, nCount); }

	/**
	 * Formatage de l'heure de fin, pour affichage
	 * \param[in]	pszBuf	Tampon recevant la chane formate
	 * \param[in]	nCount	Taille du tampon
	 * \retval		Nombre de caractres crits dans le tampon.
	 **/
	UINT32	formate_heure_fin(LPTSTR pszBuf, UINT32 nCount) const {
				return _stprintf_s(pszBuf, nCount, TEXT("%ih%02i%s"),
					fin.wHour, fin.wMinute, repetition!=0 ? TEXT("*") : TEXT("")); }

	/**
	 * Formatage de l'tat de l'enregistrement
	 * \param[in]	pszBuf	Tampon recevant la chane formate
	 * \param[in]	nCount	Taille du tampon
	 **/
	void	formate_etat(LPTSTR pszBuf, UINT32 nCount) const {
					_tcscpy_s(pszBuf, nCount, t_states[etat]); }

	int		compare_etat(EtatProgramme e2) const {
						return _tcsicmp(t_states[etat], t_states[e2]); }

	/// Obtenir le nom de la chane
	tstring	nomChaine() const;

	/// Obtenir une chane de caractres avec le nom de la programmation et le nom de la chane
	tstring	progDescr() const;
};

/**
 * \brief Classe regroupant la liste des enregistrements programms
 **/
class ProgramList : public std::vector<Programme>
{
public:
#ifdef _XMLUTILS_H_INCLUDED_
	/**
	 * Chargement de la liste des programmations  partir d'un noeud TinyXML construit
	 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	bool SetFromXMLElement(const TiXmlElement * pXmlElm);

	/**
	 * Construction d'un noeud TinyXML reprsentant la dfinition de la liste des programmations
	 * \return	Noeud XML construit
	 **/
	TiXmlElement * GetAsXMLElement() const;
#endif
};

/**
 * \brief Liste des enregistrements programms
 **/
extern ProgramList Programmes;

/**
 * \brief Traitement du chargement de la liste des programmes au dmarrage
 *
 * Inclut purge des enregistrements prims
 **/
void init_programmation();

/**
 * (Re)calculer le timer de la prochaine transition d'enregistrement
 **/
void set_timer_record();

/**
 * Trouver le temps restant avant le prochain enregistrement  effectuer.
 *
 * \param[in]		pLocaltime	Si non NULL, pointeur sur temps systme de rfrence (actuel)
 * \param[in,out]	pProg		Pointeur qui, s'il est diffrent de \p NULL, pointe sur une variable
 *								destine  recevoir l'adresse du descripteur de la programmation correspondante.
 * \param[out]		pFirstWake	Si != NULL, la variable pointe reoit le temps avant le 1er enregistrement pour
 *								lequel la sortie de mode veille a t demande, ou bien #DUREE_INFINIE si aucun
 * \return	Temps restant, exprim en millisecondes, ou bien #DUREE_INFINIE s'il n'y en a pas ou plus.
 **/
LONG temps_prochain_enregistrement(const SYSTEMTIME * pLocaltime=NULL, Programme ** pProg=NULL, LPLONG pFirstWake=NULL);

/**
 * Vrifier si le fait de quitter l'application interromprait un enregistrement en cours,
 * ou bien empcherait un enregistrement imminent.
 * \param[in]	pLocalTime	Pointeur optionnel sur le temps systme courant
 **/
bool can_power_down(const SYSTEMTIME * pLocalTime = NULL);

/**
 * Appel du dialogue de fin d'enregistrement retarde.
 **/
void delayed_stop_dialog();

/**
 * Finalisation d'un changement dans les enregistrements programms.
 **/
void finalize_prog_change();

/**
 * Fonction  appeler en cas de changement externe possible de la liste des enregistrements programms.
 * Assure la sauvegarde des modification et la mise  jour de la liste dans la fentre de dialogue
 * si celle-ci est ouverte.
 **/
void record_list_updated();

/**
 * Excution de toutes les transitions d'enregistrements (dmarrage,
 * arrt, erreur...) qui devraient tre assures  cet instant.
 *
 * \retval	\p true si au moins une transition a eu lieu dans la fonction
 **/
bool do_record_events();

/**
 * Transfert d'une description d'mission dans une structure d'enregistrement programm
 * \param[in]	canal	Rfrence sur la description de la chane
 * \param[in]	cEmis	Rfrence sur l'mission dans la liste des missions de cette chane
 * \param[out]	prog	Structure recevant la description
 **/
bool chaineToProg(const Chaine & canal, const Emission & cEmis, Programme & prog);

/**
 * Trouver l'index d'un enregistrement programm  partir de son nom
 *
 * \param[in]	nom	Nom de l'enregistrement recherch
 * \retval	 Index numrique de l'enregistrement
 * \retval	 -1 si nom vide ou rien trouv.
 **/
int trouve_prog_par_nom(LPCTSTR nom);

/**
 * Remplissage d'une combo-box de mthodes d'enregistement
 * \param[in]	hCtl	"Handle" de la combo-box
 * \param[in]	eSel	Item slectionn
 **/
void remplit_liste_methodes(HWND hCtl, MethodeEnregistrement eSel);

/**
 * Remplissage d'une combo-box d'items audio
 * \param[in]	hCtl	"Handle" de la combo-box
 * \param[in]	eSel	Item slectionn
 **/
void remplit_liste_audio(HWND hCtl, AudioMode eSel);

/**
 * Initialisation des items de mthodes d'enregistrement d'un dialogue d'enregistrements programms
 * \param[in]	hDlg	"Handle" de la bote de dialogue
 **/
void init_methodes(HWND hDlg);

/**
 * Mise  jour des mthodes d'enregistrement aprs changement d'un des items de dialogue
 * qui les contrlent
 * \param[in]	hDlg	"Handle" de la bote de dialogue
 **/
void update_methodes(HWND hDlg);

// ====================================================================================
// Dialogue des enregistrements programms
// ====================================================================================

/**
 * \brief Classe de gestion du dialogue des enregistrements programms
 **/
class CRecordDialog :
	public CModelessDialog<CRecordDialog>,
	WinSubClass<IDC_LIST_PROGRAMMES>
{
	CLViewHelper	cLVhlpr;			//!< Assistant du contrle de liste
	int				selection;			//!< Index de l'item couramment slectionn si >= 0
	NomProg			selProgNom;			/*!< Nom de la slection courante (pour pouvoir reconstituer celle-ci
											 aprs modification ou tri de la liste) */
	bool			bWarnAwake;
#if USE_TASKMAN_PASSWORD
	WCHAR			szPassword[256];	//!< Mot de passe courant
#endif

	// Variables utilises pendant les modifications des contrles date et heure pour tester la
	// diffrence entre le dbut et la fin ainsi que la diffrence avec la valeur avant changement :
	SYSTEMTIME	dtc_debut;
	SYSTEMTIME	dtc_fin;

	void update_edit_state(int selection);

#if USE_TASKMAN_PASSWORD
	/**
	 * Saisie du mot de passe requis lors de l'ajout d'une tche dans le planificateur de tches
	 **/
	bool taskLogon(LPCWSTR username);
#endif

	/**
	 * \brief Ajout de la programmation si ncessaire, incluant identification
	 *
	 * \param[in]	cProg	Programmation  ajouter
	 * \retval		\p true ou \p false selon succs ou chec, respectivement
	 **/
	bool ajoutTacheAvecLogon(Programme & cProg);

	/**
	 * Rcupre les donnes de l'enregistrement dans la fentre.
	 * Retourne 'true' si ces donnes sont valides, 'false' sinon.
	 **/
	bool ToProg(Programme & prog, bool modif);

	/// Chargement des donnes de dmarrage initialisant une nouvelle programmation
	void InitNewRecord();

	/**
	 * Transfert des donnes d'un enregistrement programm dans les items appropris de la bote
	 * de dialogue correspondante.
	 **/
	void progToDlg(const Programme & prog);

public:
	/// Constructeur
	CRecordDialog() :
		CModelessDialog(IDD_DELAYED_RECORD),
		selection(-1),
		bWarnAwake(false)
	{
		selProgNom[0] = 0;
#if USE_TASKMAN_PASSWORD
		SecureZeroMemory(szPassword, sizeof(szPassword));
#endif

		SYSTEMTIME st = {0};

		dtc_debut = dtc_fin = st;
	}

	static int CALLBACK	LVRecCompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);

	LRESULT				SubProc(WinSubClass<IDC_LIST_PROGRAMMES> & sc, UINT uMsg, WPARAM wParam, LPARAM lParam) {
							return cLVhlpr.CallWindowProc(sc, uMsg, wParam, lParam); }

	/// Traitement des messages de la bote de dialogue
	virtual INT_PTR		DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam);
};

// ====================================================================================
// Dialogue dmarrage de l'enregistrement d'une autre chane
// ====================================================================================

/**
 * \brief Classe de gestion du dialogue de dmarrage de l'enregistrement d'une autre chane
 **/
class CRecOtherDialog : public CModalDialogBase
{
	HWND	hCtlChn;
	HWND	hCtlMtd;
	HWND	hCtlAud;
	int		nChanCount;

	/// Traitement des messages de la bote de dialogue
	virtual INT_PTR DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

	/// Mise  jour de l'tat d'activation de certains items
	void UpdateStates(MethodeEnregistrement eMeth);

	/// Chargement de la liste des pistes audio d'une chane
	void FillAudioList(INT nCbIdx);

public:

	/// Constructeur :
	CRecOtherDialog() :
		CModalDialogBase(IDD_RECORD_OTHER),
		hCtlChn(NULL),
		hCtlMtd(NULL),
		hCtlAud(NULL),
		nChanCount(0)
	{
	}
};

/// Affichage d'informations concernant les prochains enregistrements
void DisplayRecordInfo(HWND hParentWnd);
